#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>

#include "rt_config.h"
#include "define.h"

#define LED_CHECK_INTERVAL 	50 /* 50ms */
#define LED_CHECK_INTERVAL_WPS 	50

#define TX_LED_BLINK_ON		30
#define TX_LED_BLINK_OFF	30

/* LED Mode */
#define LED_OFF 0
#define LED_TX_BLINK 1
#define LED_SLOW_BLINK 2
#define LED_ON 3

typedef union _LED_CFG_T
{
	struct{
#ifdef RT_BIG_ENDIAN
		UINT32 :1;
		UINT32 LED_POL:1;		/* 0: active low, 1:active high. */
		UINT32 Y_LED_MODE:2;	/* 0: off, 1: blinking upon Tx,
									2:Periodic slow blinking, 3:sloid on. */
		UINT32 G_LED_MODE:2;	/* same as Y_LED_MODE. */
		UINT32 R_LED_MODE:2;	/* same as Y_LED_MODE. */
		UINT32 :2;
		UINT32 SLOW_BLK_TIME:6;	/* slow blinking period (uint: 1sec). */
		UINT32 LED_OFF_TIIME:8;	/* Tx blinking off period (unit: 1ms). */
		UINT32 LED_ON_TIME:8; 	/* Tx blinking on period (uint: 1ms). */
#else
		UINT32 LED_ON_TIME:8; 	/* Tx blinking on period (uint: 1ms). */
		UINT32 LED_OFF_TIME:8;	/* Tx blinking off period (unit: 1ms). */
		UINT32 SLOW_BLK_TIME:6;	/* slow blinking period (uint: 1sec). */
		UINT32 :2;
		UINT32 R_LED_MODE:2;	/* same as Y_LED_MODE. */
		UINT32 G_LED_MODE:2;	/* same as Y_LED_MODE. */
		UINT32 Y_LED_MODE:2;	/* 0: off, 1: blinking upon Tx,
									2:Periodic slow blinking, 3:sloid on. */
		UINT32 LED_POL:1;		/* 0: active low, 1:active high. */
		UINT32 :1;
#endif
	} field;
	UINT32 word;
} LED_CFG_T, *PLED_CFG_T;

RALINK_TIMER_STRUCT LedCheckTimer;
RALINK_TIMER_STRUCT LedCheckTimer_wps;

unsigned char LinkStatus;
unsigned char GPIOPolarity;
unsigned char SignalStrength;

union uLedParameter LedParameter;

LED_CFG_T LedCfg;
BYTE CheckTimerEbl=0;
BYTE CheckTimerEbl_wps=0;

enum LED_STATE { SOLID_OFF=0, SOLID_ON, BLINK } led_state ;

// Entry for Led control.  Triggered by timer per 50ms
void LedCtrlMain(
	IN PVOID SystemSpecific1, 
	IN PVOID FunctionContext, 
	IN PVOID SystemSpecific2, 
	IN PVOID SystemSpecific3)
{
}


// Called when host sets LED mode.
// Called in interrupt level
void SetLinkStatus(IN PRTMP_ADAPTER pAd)
{
	switch(LinkStatus)
	{
		case 0x00:		// Radio off
		case 0x20:		// Radio on but not link up
			LedCfg.word = 0x0;
  			break;

		case 0x60:		// Radio on and link to G
		case 0xa0:		// Radio on and link to A
			LedCfg.field.LED_POL = 1;
			LedCfg.field.G_LED_MODE = LED_TX_BLINK;
			LedCfg.field.LED_ON_TIME = TX_LED_BLINK_ON;
			LedCfg.field.LED_OFF_TIME = TX_LED_BLINK_OFF;
  			break;

		case 0x10:		// WPS
			break;

		case 0x08:
			break;
	}

	RTMP_IO_WRITE32(pAd, LED_CFG, LedCfg.word);
}


void SetLinkStatusEntry(IN PRTMP_ADAPTER pAd)
{	
	SetLinkStatus(pAd);
}	// @ 0x1025

void StopLed(IN PRTMP_ADAPTER pAd)
{
	// disable WPS LED flash
	if(CheckTimerEbl_wps == 1)
	{
		BOOLEAN Cancelled;		
		RTMPCancelTimer(&LedCheckTimer_wps, &Cancelled);
		CheckTimerEbl_wps=0;
	}
	
	LedCfg.word = 0x0;
	// LED off
	RTMP_IO_WRITE32(pAd, LED_CFG, LedCfg.word);

}

void StartLed(IN PRTMP_ADAPTER pAd)
{
	SetLinkStatus(pAd);
}

void LedCtrlMain_wps(IN PVOID data)
{
	RALINK_TIMER_STRUCT *tmp = (RALINK_TIMER_STRUCT *) data;
	RTMP_ADAPTER *pAd = (RTMP_ADAPTER *) tmp->cookie;
	
	if (led_state != SOLID_OFF)
	{
		LedCfg.field.LED_POL = 0;
		LedCfg.field.G_LED_MODE = LED_OFF;
		led_state = SOLID_OFF;
	} else if (led_state != SOLID_ON)
	{
		LedCfg.field.LED_POL = 0;
		LedCfg.field.G_LED_MODE = LED_ON;
		led_state = SOLID_ON;
	}	
	RTMP_IO_WRITE32(pAd, LED_CFG, LedCfg.word);
	
	RTMPSetTimer(&LedCheckTimer_wps, LED_CHECK_INTERVAL_WPS);
}

void FlashStartLed(IN PRTMP_ADAPTER pAd)
{
	if(CheckTimerEbl_wps == 0)
	{
		RTMPInitTimer(pAd, &LedCheckTimer_wps, LedCtrlMain_wps, pAd, FALSE);
		RTMPSetTimer(&LedCheckTimer_wps, LED_CHECK_INTERVAL_WPS);
		CheckTimerEbl_wps=1;
		led_state=SOLID_OFF;
	}
}


void FlashStopLed(IN PRTMP_ADAPTER pAd)
{
	BOOLEAN Cancelled;		
	RTMPCancelTimer(&LedCheckTimer_wps, &Cancelled);
	CheckTimerEbl_wps=0;
	SetLinkStatus(pAd);
}		

