
#include <rt_config.h>


void very_simple_derange(unsigned char *buf, int size, int arrange)
{
    unsigned short utmp;
    int i0, j0, k0;

    i0 = -1, j0 = size;

    while (++i0 < --j0) 
	{
        utmp = buf[i0] | (buf[j0]<<8);

        k0 = i0 & 0x0F;
        if (arrange) k0 = 16 - k0;
        utmp = ~((utmp >> k0) | (utmp << (16-k0)));

        buf[i0] = (unsigned char)utmp;
        buf[j0] = (unsigned char)(utmp>>8);
    }
}

static uint32_t rand_r(uint32_t *seed)
{
    uint32_t result, next = *seed;

    next *= 1103515245;
    next += 12345;
    result = (next / 65536) % 2048;

    next *= 1103515245;
    next += 12345;
    result <<= 10;
    result ^= (next / 65536) % 1024;

    next *= 1103515245;
    next += 12345;
    result <<= 10;
    result ^= (next / 65536) % 1024;

    *seed = next;

    return result;
}

void jse_encode(uint32_t *src, uint32_t *enc, int size, uint32_t seed)
{
    uint32_t k,d;
    int i;

    k = rand_r(&seed);
    for (i = 0; i < size; i++)
    {
        d = *enc++ = *src++ ^ k;
        k = d ^ rand_r(&seed);
    }
}

void jse_decode(uint32_t *src, uint32_t *dec, int size, uint32_t seed)
{
    uint32_t k, d;
    int i, j;

    k = rand_r(&seed);
    for(i = 0; i< size; i++)
    {
        j = *src++;
        d = *dec++ = j ^ k;
        k = j ^ rand_r(&seed);
    }
}

#define ALG_VSD				0x80
#define ALG_JSE         	0x81

int rds_arrange(unsigned char *uh, int start, int end)
{
	uint16_t utmp;
	uint32_t seed;

	int alg = uh[start+0];
	// int opt = uh[start+1];

	switch (alg) 
	{
		case ALG_VSD:
        //only 0x00 is allowed
        // if (opt != 0) return -1;
        very_simple_derange(uh+start+2, end-start-2, 1);
        break;

		case ALG_JSE:
		//start to decode (arange)
		seed = uh[start+2] * 256 + uh[start+3];
		jse_decode((uint32_t*)(uh+start+4), (uint32_t*)(uh+start+4), (end-start-4)/4, seed);
		utmp = *(uint16_t*)(uh+start+2);
		*(uint16_t*)(uh+start+2) = *(uint16_t*)(uh+end-2);
		*(uint16_t*)(uh+end-2) = utmp;
	}

	uh[start+0] = uh[13];
	uh[start+1] = uh[27];
	uh[13] = uh[27] = 0;

	return 0;
}

int rds_derange(unsigned char *uh, int alg, int opt, int start, int end)
{
	uint16_t utmp;
	uint32_t seed;

	uh[13] = uh[start+0];
	uh[27] = uh[start+1];

	//we only support RDS/00036 arrangement
	if (alg != ALG_JSE)
	{
		//DERANGE_ALG is not supported
		return -1;
	}

	//start to encode (derange)
	utmp = *(uint16_t*)(uh+start+2);
	*(uint16_t*)(uh+start+2) = *(uint16_t*)(uh+end-2);
	*(uint16_t*)(uh+end-2) = utmp;
	seed = uh[start+2] * 256 + uh[start+3];
	jse_encode((uint32_t*)(uh+start+4), (uint32_t*)(uh+start+4), (end-start-4)/4, seed);

	uh[start+0] = alg;
	uh[start+1] = opt;

	return 0;
}

