#if defined(WLAN_LED)

#ifdef LINUX
#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#endif // LINUX //

#include "rt_config.h"
#include "rt_led.h"

#define LED_GPIO CONFIG_RT2860V2_LED_GPIO

#define LED_CHECK_INTERVAL 		500 /* 50ms */
#define LED_CHECK_INTERVAL_WPS 	50 /* 1ms */

#define RALINK_PIO_BASE                 0xB0000600
#define RALINK_PRGIO_ADDR		RALINK_PIO_BASE // Programmable I/O
#define RALINK_REG_PIODATA							(RALINK_PRGIO_ADDR + 0x20)
#define RALINK_REG_PIO3924DATA          (RALINK_PRGIO_ADDR + 0x48)
#define RALINK_REG_PIO5140DATA          (RALINK_PRGIO_ADDR + 0x70)
#define RALINK_REG_PIO7140DATA          (RALINK_PRGIO_ADDR + 0x70)
#define RALINK_REG_PIO9572DATA          (RALINK_PRGIO_ADDR + 0x98)

#define RALINK_GPIO_0			0x00000001
#define RALINK_GPIO_1			0x00000002
#define RALINK_GPIO_2			0x00000004
#define RALINK_GPIO_3			0x00000008
#define RALINK_GPIO_4			0x00000010
#define RALINK_GPIO_5			0x00000020
#define RALINK_GPIO_6			0x00000040
#define RALINK_GPIO_7			0x00000080
#define RALINK_GPIO_8			0x00000100
#define RALINK_GPIO_9			0x00000200
#define RALINK_GPIO_10			0x00000400
#define RALINK_GPIO_11			0x00000800
#define RALINK_GPIO_12			0x00001000
#define RALINK_GPIO_13			0x00002000
#define RALINK_GPIO_14			0x00004000
#define RALINK_GPIO_15			0x00008000
#define RALINK_GPIO_16			0x00010000
#define RALINK_GPIO_17			0x00020000
#define RALINK_GPIO_18			0x00040000
#define RALINK_GPIO_19			0x00080000
#define RALINK_GPIO_20			0x00100000
#define RALINK_GPIO_21			0x00200000
#define RALINK_GPIO_22			0x00400000
#define RALINK_GPIO_23			0x00800000
#define RALINK_GPIO(x)			(1 << x)


#define __LED_ON(gpio) ralink_gpio_led_value |= RALINK_GPIO(gpio);
#define __LED_OFF(gpio) ralink_gpio_led_value &= ~RALINK_GPIO(gpio);

u32 ralink_gpio_led_value = 0;


RALINK_TIMER_STRUCT LedCheckTimer;
RALINK_TIMER_STRUCT LedCheckTimer_wps;
typedef struct {
        int gpio;                       //gpio number (0 ~ 23)
        unsigned int on;                //interval of led on
        unsigned int off;               //interval of led off
        unsigned int blinks;            //number of blinking cycles
        unsigned int rests;             //number of break cycles
        unsigned int times;             //blinking times
} ralink_gpio_led_info;

void SetLinkStatus(IN PRTMP_ADAPTER pAd);


unsigned char LinkStatus;
unsigned char SignalStrength;
unsigned char GPIOPolarity;
ULED_PARAMETER LedParameter;

int ralink_gpio_led_set(ralink_gpio_led_info led);

BYTE CheckTimerEbl=0;
BYTE CheckTimerEbl_wps=0;

enum LED_STATE { SOLID_OFF=0, SOLID_ON, BLINK } led_state ;

void SetLinkStatusEntry(IN PRTMP_ADAPTER pAd)
{
        SetLinkStatus(pAd);
}       // @ 0x1025

// Called when host sets LED mode.
// Called in interrupt level
void SetLinkStatus(IN PRTMP_ADAPTER pAd)
{	
	if(CheckTimerEbl==0){
		memset(&LedCheckTimer, 0, sizeof(RALINK_TIMER_STRUCT));
                RTMPInitTimer(pAd, &LedCheckTimer,GET_TIMER_FUNCTION(LedCtrlMain), pAd, TRUE);
                RTMPSetTimer(&LedCheckTimer, LED_CHECK_INTERVAL);
                CheckTimerEbl=1;
		led_state=SOLID_OFF ;
	}	
	//printk(KERN_INFO "%s : LinkStatus=%02x \n", __FUNCTION__, LinkStatus);			
}

static inline int TX_TRAFFIC_EXIST(
        IN PRTMP_ADAPTER pAd)
{
        UINT RegValue;

        RTMP_IO_READ32(pAd, MCU_INT_STATUS, &RegValue);
        return (RegValue & 0x0e);
}

// Entry for Led control.  Triggered by timer per 50ms
void LedCtrlMain(
	IN PVOID SystemSpecific1,
        IN PVOID FunctionContext,
        IN PVOID SystemSpecific2,
        IN PVOID SystemSpecific3)
{
	RTMP_ADAPTER *pAd = (RTMP_ADAPTER *)FunctionContext;
	ralink_gpio_led_info led;	
	if (TX_TRAFFIC_EXIST(pAd)) // Check if there is Tx Traffic
	{
		if( led_state != BLINK ){
			led_state = BLINK;
			led.gpio=LED_GPIO;
                	led.on=1;
                	led.off=1;
                	led.blinks=4000;
                	led.rests=0;
                	led.times=4000;
                	ralink_gpio_led_set(led);	
		}	
	}else{
		
		if( led_state != SOLID_ON){
			led_state = SOLID_ON;
			led.gpio=LED_GPIO;
                        led.on=4000;
                        led.off=0;
                        led.blinks=1;
                        led.rests=0;
                        led.times=4000;
                        ralink_gpio_led_set(led);
		}
	}
	// Clear Tx and beacon Tx complete interrupt
        RTMP_IO_WRITE32(pAd, MCU_INT_STATUS, 0x1e);	
	RTMPSetTimer(&LedCheckTimer, LED_CHECK_INTERVAL);
}

void StopLed(IN PRTMP_ADAPTER pAd)
{
	BOOLEAN Cancelled;	
	ralink_gpio_led_info led;	
	RTMPCancelTimer(&LedCheckTimer, &Cancelled);	
	//Led Off
	led.gpio=LED_GPIO;
        led.on=0;
       	led.off=4000;
       	led.blinks=0;
      	led.rests=1;
      	led.times=4000;
    	ralink_gpio_led_set(led);
	CheckTimerEbl=0;
}

void StartLed(IN PRTMP_ADAPTER pAd)
{
	SetLinkStatus(pAd);
}



void FlashStartLed(IN PRTMP_ADAPTER pAd)
{
	//printk(KERN_INFO "%s : LinkStatus=wuming \n", __FUNCTION__);
		
	if(CheckTimerEbl_wps==0){
		memset(&LedCheckTimer_wps, 0, sizeof(RALINK_TIMER_STRUCT));
    RTMPInitTimer(pAd, &LedCheckTimer_wps, GET_TIMER_FUNCTION(LedCtrlMain_wps), pAd, TRUE);
    RTMPSetTimer(&LedCheckTimer_wps, LED_CHECK_INTERVAL);
    CheckTimerEbl_wps=1;
		led_state=SOLID_OFF ;
	}
}


void FlashStopLed(IN PRTMP_ADAPTER pAd)
{
	BOOLEAN Cancelled;
	ralink_gpio_led_info led;		
	RTMPCancelTimer(&LedCheckTimer_wps, &Cancelled);
	//Led Off
	led.gpio=LED_GPIO;
        led.on=0;
        led.off=4000;
        led.blinks=0;
        led.rests=1;
        led.times=4000;
        ralink_gpio_led_set(led);
	CheckTimerEbl_wps=0;
}		

void LedCtrlMain_wps(
	IN PVOID SystemSpecific1,
        IN PVOID FunctionContext,
        IN PVOID SystemSpecific2,
        IN PVOID SystemSpecific3)
{
	if( led_state != SOLID_OFF )
	{
		
		if(LED_GPIO<=23)
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIODATA);
			__LED_OFF(LED_GPIO);
			*(volatile u32 *)(RALINK_REG_PIODATA) = ralink_gpio_led_value;
		}
		else if(LED_GPIO<=39)
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIO3924DATA);
			__LED_OFF(LED_GPIO-24);			
			*(volatile u32 *)(RALINK_REG_PIO3924DATA) = ralink_gpio_led_value;
		}
		else if(LED_GPIO<=39)
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIO3924DATA);
			__LED_OFF(LED_GPIO-24);			
			*(volatile u32 *)(RALINK_REG_PIO3924DATA) = ralink_gpio_led_value;
		}
#ifdef CONFIG_RALINK_RT3883
		else if(LED_GPIO<=71)
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIO7140DATA);
			__LED_OFF(LED_GPIO-40);			
			*(volatile u32 *)(RALINK_REG_PIO7140DATA) = ralink_gpio_led_value;
		}
		else
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIO9572DATA);
			__LED_OFF(LED_GPIO-72);			
			*(volatile u32 *)(RALINK_REG_PIO9572DATA) = ralink_gpio_led_value;			
		}
#else
		else
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIO5140DATA);
			__LED_OFF(LED_GPIO-40);			
			*(volatile u32 *)(RALINK_REG_PIO5140DATA) = ralink_gpio_led_value;			
		}
#endif			
		
 		//printk(KERN_INFO "%s :SOLID_OFF ralink_gpio_led_value = 0x%x  \n", __FUNCTION__, ralink_gpio_led_value);
		led_state = SOLID_OFF;
	}	
	else if( led_state != SOLID_ON)
	{		
		if(LED_GPIO<=23)
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIODATA);
			__LED_ON(LED_GPIO);
			*(volatile u32 *)(RALINK_REG_PIODATA) = ralink_gpio_led_value;			
		}
		else if(LED_GPIO<=39)
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIO3924DATA);
			__LED_ON(LED_GPIO-24);			
			*(volatile u32 *)(RALINK_REG_PIO3924DATA) = ralink_gpio_led_value;
		}
		else if(LED_GPIO<=39)
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIO3924DATA);
			__LED_ON(LED_GPIO-24);			
			*(volatile u32 *)(RALINK_REG_PIO3924DATA) = ralink_gpio_led_value;
		}
#ifdef CONFIG_RALINK_RT3883
		else if(LED_GPIO<=71)
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIO7140DATA);
			__LED_ON(LED_GPIO-40);			
			*(volatile u32 *)(RALINK_REG_PIO7140DATA) = ralink_gpio_led_value;
		}
		else
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIO9572DATA);
			__LED_ON(LED_GPIO-72);			
			*(volatile u32 *)(RALINK_REG_PIO9572DATA) = ralink_gpio_led_value;			
		}
#else
		else
		{
			ralink_gpio_led_value = *(volatile u32 *)(RALINK_REG_PIO5140DATA);
			__LED_ON(LED_GPIO-40);			
			*(volatile u32 *)(RALINK_REG_PIO5140DATA) = ralink_gpio_led_value;			
		}
#endif
				
		led_state = SOLID_ON;
	}
	RTMPSetTimer(&LedCheckTimer_wps, LED_CHECK_INTERVAL_WPS);
}
#endif
