/****************************************************************************
* File:         $Workfile$
* Revision:     $Revision: 1.2.4.1 $
* Date:         $Date: 2009-10-21 05:50:12 $
*
* Overview:     
* Functions:
*
* Author:       Ian Tsao
* 
* -- Metanoia Copyright Notice --
*
* (C) Copyright 2005 Metanoia, Inc. All rights reserved.
* 
* Metanoia reserves the right to change specifications without notice. 
* Copyright 2005 Metanoia, Inc. All rights reserved. Preliminary Data-Sheet 
* indicates this product is in design and the specification may change. 
* Electrical parametrics have not been analyzed and are not specified. Do not 
* use this data sheet as a design reference document. Please contact Metanoia 
* for a current data sheet that may be used as a design reference. All 
* contents of this document are protected by copyright law and may not be 
* reproduced without the express written consent of Metanoia, Inc. Metanoia, 
* the Metanoia logo, and combinations thereof are trademarks of Metanoia, Inc. 
* Other product names used in this publication are for identification purposes 
* only and may be trademarks or registered trademarks of their respective 
* companies. The contents of this document are provided in connection with 
* Metanoia, Inc. products. Metanoia, Inc. has made best efforts to ensure that 
* the information contained herein is accurate and reliable. However, Metanoia, 
* Inc. makes no warranties, express or implied, as to the accuracy or 
* completeness of the contents of this publication and is providing this 
* publication 'as is'. Metanoia, Inc. reserves the right to make changes to 
* specifications and product descriptions at any time without notice, and to 
* discontinue or make changes to its products at any time without notice. 
* Metanoia, Inc. does not assume any liability arising out of the application 
* or use of any product or circuit, and specifically disclaims any and all 
* liability, including without limitation special, consequential, or 
* incidental damages.
* 
****************************************************************************/
#ifndef _AWARDM_H_
#define _AWARDM_H_


// bus width DMT address shift
#define BUS_WIDTH_8_SHIFT 16
#define BUS_WIDTH_16_SHIFT 17
#define BUS_WIDTH_32_SHIFT 18


#ifdef AWARD

// GPIO_EN is the address at which controller enables/disables GPIO pin
#define GPIO_EN 0x40000010

// GPIO_BEN is the address at which controller sets GPIO bit enable
#define GPIO_BEN 0x40700004
// GPIO_DIR is the address at which controller sets GPIO directions
#define GPIO_DIR 0x40700008

// GPIO_DATA is the address at which controller accesses other Falcon
// interface pins (IRQA-D, RST_N, ...)
#define GPIO_DATA 0x40700000

#define EBI_RDY 0x40100028
#define EBI_BCR2 0x40100008

// Eth reset
#define ETH_RESET 0x40000000

// FALCON_PORT is the address at which controller sees HPI port
#define FALCON_PORT 0x22000000

// GPIO bit # for RESP_RDY access
#define RESP_RDY_BIT 0x1000000 
// IRQA-D lines are connected to bits 21-18 of GTX GPIO port
#define BOOT_MODE 0x1c0000	// Falcon is in HPI slave boot mode
// Cfg register for memory bank #2, connected to falcon
// FALCON_MASK - these are all GPIO bits connected to Falcon
#define FALCON_MASK 0x3f0000
// GPIO bit connected to Falcon reset
#define RST_N 0x10000
// GPIO bit # for IRQA access
#define IRQA_BIT 0x40000 
// switch PIN Setting
#define SMI_EECS 0x00000020
#define SMI_EESK 0x00000010
#define SMI_EDI  0x00000080
#define SMI_EDO  0x00000040		
	

#elif defined (XENOS)


// GPIO_EN is the address at which controller enables/disables GPIO pin
#define GPIO_EN 0x40000010

// GPIO_BEN is the address at which controller sets GPIO bit enable
#define GPIO_BEN 0x40700004
// GPIO_DIR is the address at which controller sets GPIO directions
#define GPIO_DIR 0x40700008

// GPIO_DATA is the address at which controller accesses other Falcon
// interface pins (IRQA-D, RST_N, ...)
#define GPIO_DATA 0x40700000

#define EBI_RDY 0x40100028
#define EBI_BCR2 0x40100008

// Eth reset
#define ETH_RESET 0x40000000

// XENOS_PORT is the address at which controller sees HPI port
#ifndef VIA_SAMSUNG
#define XENOS_PORT 0x18000000
#else
#define XENOS_PORT 0x82000000
#endif

// GPIO bit # for RESP_RDY access
#define RESP_RDY_BIT 0x1000000 
// IRQA-D lines are connected to bits 21-18 of GTX GPIO port
#define BOOT_MODE 0x1c0000	// Falcon is in HPI slave boot mode
// Cfg register for memory bank #2, connected to falcon
// FALCON_MASK - these are all GPIO bits connected to Falcon
#define FALCON_MASK 0x3f0000
// GPIO bit connected to Falcon reset
#define RST_N 0x10000
// GPIO bit # for IRQA access
#define IRQA_BIT 0x40000 
// switch PIN Setting
#define SMI_EECS 0x00000020
#define SMI_EESK 0x00000010
#define SMI_EDI  0x00000080
#define SMI_EDO  0x00000040		


#elif defined (BOARD_A3)

// GPIO_EN is the address at which controller enables/disables GPIO pin
#define GPIO_EN 0x40000010

// GPIO_BEN is the address at which controller sets GPIO bit enable
#define GPIO_BEN 0x40700004
// GPIO_DIR is the address at which controller sets GPIO directions
#define GPIO_DIR 0x40700008

// GPIO_DATA is the address at which controller accesses other Falcon
// interface pins (IRQA-D, RST_N, ...)
#define GPIO_DATA 0x40700000

#define EBI_RDY 0x40100028
#define EBI_BCR2 0x40100008

// Eth reset
#define ETH_RESET 0x40000000

#ifdef AWARD
	// FALCON_PORT is the address at which controller sees HPI port
	#define FALCON_PORT 0x22000000
#endif

   
// IRQA-D lines are connected to bits 21-18 of GTX GPIO port
#define BOOT_MODE 0x0A000200	// Falcon is in HPI slave boot mode
// Cfg register for memory bank #2, connected to falcon
// FALCON_MASK - these are all GPIO bits connected to Falcon
#define FALCON_MASK 0x1E000F0F
// GPIO bit connected to Falcon reset
#define RST_N 0x800
// GPIO bit # for IRQA access
#define IRQA_BIT 0x200 
// GPIO bit # for RESP_RDY access
#define RESP_RDY_BIT 0x4000000 	
// switch PIN Setting

#define SMI_EECS 0x00000002
#define SMI_EESK 0x00000004
#define SMI_EDI  0x00000001
#define SMI_EDO  0x00000008	

// switch register definition
#define SMI_ADDRESS_TABLE_CONTROL0 0x11A
#define SMI_ADDRESS_TABLE_CONTROL1 0x11B
#define SMI_ADDRESS_TABLE_CONTROL2 0x11C
#define SMI_ADDRESS_TABLE_CONTROL3 0x11D
#define SMI_ADDRESS_TABLE_CONTROL4 0x11E
#define SMI_ADDRESS_TABLE_CONTROL5 0x11F

#define SMI_ADDRESS_TABLE_STATUS0 0x120						
#define SMI_ADDRESS_TABLE_STATUS1 0x121						
#define SMI_ADDRESS_TABLE_STATUS2 0x122						
#define SMI_ADDRESS_TABLE_STATUS3 0x123						
#define SMI_ADDRESS_TABLE_STATUS4 0x124						
#define SMI_ADDRESS_TABLE_STATUS5 0x125												


#elif defined (HOPE)


// GPIO_EN is the address at which controller enables/disables GPIO pin
#define GPIO_EN 0x40000010

// GPIO_BEN is the address at which controller sets GPIO bit enable
#define GPIO_BEN 0x40700004
// GPIO_DIR is the address at which controller sets GPIO directions
#define GPIO_DIR 0x40700008

// GPIO_DATA is the address at which controller accesses other Falcon
// interface pins (IRQA-D, RST_N, ...)
#define GPIO_DATA 0x40700000

#define EBI_RDY 0x40100028
#define EBI_BCR2 0x40100008

// Eth reset
#define ETH_RESET 0x40000000

// FALCON_PORT is the address at which controller sees HPI port
#ifndef VIA_SAMSUNG
#define FALCON_PORT 0x18000000
#else
#define FALCON_PORT 0x82000000
#endif

// GPIO bit # for RESP_RDY access
#define RESP_RDY_BIT 0x1000000 
// IRQA-D lines are connected to bits 21-18 of GTX GPIO port
#define BOOT_MODE 0x1c0000	// Falcon is in HPI slave boot mode
// Cfg register for memory bank #2, connected to falcon
// FALCON_MASK - these are all GPIO bits connected to Falcon
#define FALCON_MASK 0x3f0000
// GPIO bit connected to Falcon reset
#define RST_N 0x10000
// GPIO bit # for IRQA access
#define IRQA_BIT 0x40000 
// switch PIN Setting
#define SMI_EECS 0x00000020
#define SMI_EESK 0x00000010
#define SMI_EDI  0x00000080
#define SMI_EDO  0x00000040		


#elif defined (BOARD_A5)

// GPIO_EN is the address at which controller enables/disables GPIO pin
#define GPIO_EN 0x40000010

// GPIO_BEN is the address at which controller sets GPIO bit enable
#define GPIO_BEN 0x40700004
// GPIO_DIR is the address at which controller sets GPIO directions
#define GPIO_DIR 0x7c000008

// GPIO_DATA is the address at which controller accesses other Falcon
// interface pins (IRQA-D, RST_N, ...)
#define GPIO_DATA  			0x7c000000
#define GPIO_DATA_INPUT     	0x7c000004

#define EBI_RDY 0x40100028
#define EBI_BCR2 0x40100008

// Eth reset
#define ETH_RESET 0x40000000

// FALCON_PORT is the address at which controller sees HPI port
#if defined (RALINK_RT3052) 
//#define FALCON_PORT 0x1B000000
#define FALCON_PORT 0x1B000100
#elif defined (ICPLUS_IP3210)
#define FALCON_PORT 0xBF000000
#elif defined (STAR_STR9105)
#define FALCON_PORT 0x18800000
#endif

// GPIO bit # for RESP_RDY access
#define RESP_RDY_BIT 0x1000000 
// IRQA-D lines are connected to bits 21-18 of GTX GPIO port
#define BOOT_MODE 0x1c0000	// Falcon is in HPI slave boot mode
// Cfg register for memory bank #2, connected to falcon
// FALCON_MASK - these are all GPIO bits connected to Falcon
#define FALCON_MASK 0x3f0000
// GPIO bit connected to Falcon reset
#define RST_N 0x10000
// GPIO bit # for IRQA access
#define IRQA_BIT 0x40000 
// switch PIN Setting
#define SMI_EECS (0x1<<14)
#define SMI_EESK (0x1<<15)
#define SMI_EDI   (0x1<<17)
#define SMI_EDO  (0x1<<16)


#endif

#define BIT_31                          0x80000000
#define BIT_30                          0x40000000
#define BIT_29                          0x20000000
#define BIT_28                          0x10000000
#define BIT_27                          0x08000000
#define BIT_26                          0x04000000
#define BIT_25                          0x02000000
#define BIT_24                          0x01000000
#define BIT_23                          0x00800000
#define BIT_22                          0x00400000
#define BIT_21                          0x00200000
#define BIT_20                          0x00100000
#define BIT_19                          0x00080000
#define BIT_18                          0x00040000
#define BIT_17                          0x00020000
#define BIT_16                          0x00010000
#define BIT_15                          0x00008000
#define BIT_14                          0x00004000
#define BIT_13                          0x00002000
#define BIT_12                          0x00001000
#define BIT_11                          0x00000800
#define BIT_10                          0x00000400
#define BIT_9                           0x00000200
#define BIT_8                           0x00000100
#define BIT_7                           0x00000080
#define BIT_6                           0x00000040
#define BIT_5                           0x00000020
#define BIT_4                           0x00000010
#define BIT_3                           0x00000008
#define BIT_2                           0x00000004
#define BIT_1                           0x00000002
#define BIT_0                           0x00000001

extern mt_uint32 get_max_chips(void);
extern mt_uint32 init_system_board(void);
extern mt_uint32 check_response_rdy(void);
extern mt_uint8 write_hpi_8(mt_uint8 chipid, mt_uint8 data);
extern mt_uint8 read_hpi_8(mt_uint8 chipid);
extern mt_uint32 reset_chip(mt_uint32 chipid);
extern mt_uint32 reset_chip_high(mt_uint32 chipid);
extern mt_uint32 reset_chip_low(mt_uint32 chipid);
extern void trigger_irq(void);
extern void set_rdy_pin(void);

#endif	//_AWARDM_H_
