/****************************************************************************
* File:         $Workfile$
* Revision:     $Revision: 1.2.8.1 $
* Date:         $Date: 2010-08-10 06:38:15 $
*
* Overview:     
* Functions:
*
* Author:       Ian Tsao
* 
* -- Metanoia Copyright Notice --
*
* (C) Copyright 2005 Metanoia, Inc. All rights reserved.
* 
* Metanoia reserves the right to change specifications without notice. 
* Copyright 2005 Metanoia, Inc. All rights reserved. Preliminary Data-Sheet 
* indicates this product is in design and the specification may change. 
* Electrical parametrics have not been analyzed and are not specified. Do not 
* use this data sheet as a design reference document. Please contact Metanoia 
* for a current data sheet that may be used as a design reference. All 
* contents of this document are protected by copyright law and may not be 
* reproduced without the express written consent of Metanoia, Inc. Metanoia, 
* the Metanoia logo, and combinations thereof are trademarks of Metanoia, Inc. 
* Other product names used in this publication are for identification purposes 
* only and may be trademarks or registered trademarks of their respective 
* companies. The contents of this document are provided in connection with 
* Metanoia, Inc. products. Metanoia, Inc. has made best efforts to ensure that 
* the information contained herein is accurate and reliable. However, Metanoia, 
* Inc. makes no warranties, express or implied, as to the accuracy or 
* completeness of the contents of this publication and is providing this 
* publication 'as is'. Metanoia, Inc. reserves the right to make changes to 
* specifications and product descriptions at any time without notice, and to 
* discontinue or make changes to its products at any time without notice. 
* Metanoia, Inc. does not assume any liability arising out of the application 
* or use of any product or circuit, and specifically disclaims any and all 
* liability, including without limitation special, consequential, or 
* incidental damages.
* 
****************************************************************************/
#include "dsll2.h"
#include "ioctlmeta.h"

#ifdef __KERNEL__
//semaphore used to control hpi r/w throhgh IOCTL
//static DECLARE_MUTEX(mr_sem_chip);
#endif


void set_status_idle_chip(mt_uint32 chipid)
{
	mt_uint32 chip;

	if(chipid == MT_PORT_BroadCast) {
		for(chip = 0; chip < MAX_CHIPS; chip++)
			port_info[chip].status = MT_OPSTATE_IDLE;
	}
	else
		port_info[chipid].status = MT_OPSTATE_IDLE;
}


mt_uint32 download_chip(mt_uint32 chipid,mt_uint32 bufsize,mt_uint8* buffer)
{
	mt_uint32 i;
	mt_uint8 ret;
	int j;

	for(i = 0; i < bufsize; i++) {
		ret = write_hpi_8(chipid, buffer[i]);
		if(ret != MT_OK) return ret;

		if(i%100 == 0) PRINTL(DBG_LVL_CHIP, ".");

		for(j=0;j<1000;j++);
	}

	return MT_OK;	
}


mt_uint32 boot_chip(mt_uint32 chipid, mt_uint32 firmwareid)
{
	fw_t *pfw;
	mt_uint8 *p_code, ret;
	mt_uint32 i;
	mt_uint8 chip;
	mt_uint32 j;

	PRINTL(DBG_LVL_CHIP,"boot_chip: chipid=%i, firmwareid=%i\n", chipid, firmwareid);

	if(chipid == MT_PORT_BroadCast) {
		for(chip = 0; chip < MAX_CHIPS; chip++) {
			if(chips[chip].fw == NULL)
				return MT_FW_NOT_SET_UP;
		}

		pfw = chips[0].fw;	//all port 0-23 point to the same fw, so just pick anyone of them
	}
	else {
		if(chips[chipid].fw == NULL)
			return MT_FW_NOT_SET_UP;

		pfw = chips[chipid].fw;
	}

	if(pfw[firmwareid].code_size == 0)
		return MT_NO_CODE_TO_LOAD;

	PRINTL(DBG_LVL_CHIP,"firmware size=%i\n",pfw[firmwareid].code_size);
	p_code = pfw[firmwareid].code;

	for(i = 0; i < pfw[firmwareid].code_size; i++) {
		ret = write_hpi_8(chipid, p_code[i]);
		if(ret != MT_OK) return ret;

		if(i%100 == 0) PRINTL(DBG_LVL_CHIP, ".");

		for(j=0;j<1000;j++);
	}
	PRINTL(DBG_LVL_CHIP, "\n");

#ifdef XENOS
	//delay 1ms for waiting firmware download complete
	for(i=0;i<200000;i++);
#endif
#ifdef HOPE
	//delay 1ms for waiting firmware download complete
	for(i=0;i<200000;i++);
#endif
#ifdef BOARD_A5
	//delay 1ms for waiting firmware download complete
	for(i=0;i<200000;i++);
#endif

	set_status_idle_chip(chipid);

	return MT_OK;
}


mt_uint32 init_chip_chip(mt_uint32 chipid, mt_uint32 firmwareid)
{
	mt_uint8 chip;
#ifndef AWARD
	mt_uint32 i;
#endif

	PRINTL(DBG_LVL_CHIP, "init_chip_chip:\n");

	if(chipid == MT_PORT_BroadCast) {
		for(chip = 0; chip < MAX_CHIPS; chip++) {
			if(chips[chip].fw == NULL)
				load_firmware(&chips[chip].fw);		
		}
	}
	else {
		if(chips[chipid].fw == NULL)
			load_firmware(&chips[chipid].fw);		
	}

	reset_chip(chipid);

#ifndef AWARD
	//delay for waiting reset complete
	for(i=0;i<25000000;i++);
#endif

	return boot_chip(chipid, firmwareid);
}


/*mt_uint32 init_system_chip(mt_uint32 firmwareid)
{
	mt_uint32 i, result;

	PRINTL(DBG_LVL_CHIP,"init_system_chip:\n");
#ifdef AWARD	
	result = init_system_board();
	if(result != MT_OK) {
		PRINTL(DBG_LVL_CHIP,"init_system_chip: InitSystem_Board failed - %i.\n", result);
 		return result;
	}
#endif
	result = init_chip_chip(MT_PORT_BroadCast, firmwareid);

	for(i=0; i<MAX_CHIPS; i++) {
		set_mapper_ports(i, chips[i].fw[firmwareid].ports);
	}

	return result;
}*/


mt_uint32 handle_rx_frame_chip(mt_uint32 chipid)
{
	mt_uint8 *pbuf;
	mt_uint32 len, count;

	pbuf = &chips[chipid].buf[0];
 	count = len = (pbuf[0] << 8 ) | pbuf[1];
 	
 	PRINTL(DBG_LVL_CHIP,"handle_rx_frame_chip:%i\n",count);
 	
	if((count<MT_MAX_BUFFERSIZE) && ((signed int)count>0) && (debug_level & DBG_LVL_CHIP)) {
		PRINTL(DBG_LVL_CHIP,"Chip: ");
		while(count--)
			PRINTL(DBG_LVL_CHIP,"%02x ", *pbuf++);
		PRINTL(DBG_LVL_CHIP,"\n");
	}
	PRINTL(DBG_LVL_CHIP,"handle_rx_frame_chip:end print,count=%i\n",count);
	pbuf = &chips[chipid].buf[0] + 2;
	return handle_rx_frame_mapper(chipid, pbuf, len-2);
}


mt_uint32 new_frame_exchange(mt_uint32 chipid, mt_uint32 len, mt_uint8 *pframe)
{
	int i;
	mt_uint8 *pbuf;
	//mt_uint32 count;
		
	mt_uint32 llen = len + 2;
	PRINTL(DBG_LVL_CHIP,"new_frame_exchange:\n");
	if(llen > MT_MAX_BUFFERSIZE) {
		PRINTL(DBG_LVL_CHIP,"new_frame_exchange: Frame too long - %i\n", llen);
	}
	if(llen < 3) {
		PRINTL(DBG_LVL_CHIP,"new_frame_exchange: Frame too short - %i\n", llen);
	}

/*#ifdef __KERNEL__
//semaphore control
	down_interruptible(&mr_sem_chip);
#endif*/

#ifdef AWARD
	// Trigger the interrupt in Falcon
	trigger_irq();
#endif
#ifdef BOARD_A5
	// Trigger the interrupt in Falcon
	trigger_irq();
#endif

	//delay for gap of two transactions and also for IRQ trigger delay
	for(i=0;i<trans_delay;i++);

	write_hpi_8(chipid, (llen >> 8 ) & 0xff);
	for(i=0;i<w_delay;i++);
	write_hpi_8(chipid, llen & 0xff);
	for(i=0;i<w_delay;i++);

	//printk("%02x ", (llen >> 8 ) & 0xff);
	//printk("%02x ", llen & 0xff);

	//printk("%02x ", *pframe);
	write_hpi_8(chipid, *pframe++);
	len--;
	for(i=0;i<w_delay;i++);

	//for(i=0;i<hdr_delay;i++);

	while(len--) {
		//printk("%02x ", *pframe);
		write_hpi_8(chipid, *pframe++);
		for(i=0;i<w_delay;i++);
	}
	//printk("\n");
	
	if(manuallen) {
		for(manuallen-=llen;manuallen>0;manuallen--) {
			write_hpi_8(chipid, (mt_uint8)0x00);
		}
	}

	for(i=0;i<rw_delay;i++);

	/*
	count = 0x10000000;
	while(count--) {
		if(!check_response_rdy()) {
			PRINTD("CheckResponseRDY break\n");
			break;
		}
	}
	PRINTD("CheckResponseRDY Out\n");    
   	*/
   	
	if(readdsp) {
		pbuf = &chips[chipid].buf[0];
		memset(pbuf, 0, sizeof(chips[chipid].buf));
		pbuf[0] = read_hpi_8(chipid);
		for(i=0;i<r_delay;i++);
		pbuf[1] = read_hpi_8(chipid);
		for(i=0;i<r_delay;i++);
		//printk("%02x ", pbuf[0]);
		//printk("%02x ", pbuf[1]);
	 	llen = (pbuf[0] << 8 ) | pbuf[1];

		if(llen > MT_MAX_BUFFERSIZE) {
			PRINTD("new_frame_exchange: HDLC frame too long - %i\n", llen);
			llen = MT_MAX_BUFFERSIZE;
	 		pbuf[0] = (llen >> 8) & 0x000000ff;
	 		pbuf[1] = llen & 0x000000ff;
		}

		pbuf += 2;
		while(llen-- > 2) {
			*pbuf = read_hpi_8(chipid);
			//printk("%02x ", *pbuf);
			pbuf++;
			for(i=0;i<r_delay;i++);
		}
		//printk("\n");
	}

/*#ifdef __KERNEL__
//semaphore control
	up(&mr_sem_chip);
#endif*/

	return handle_rx_frame_chip(chipid);
}


mt_uint32 frame_exchange(mt_uint32 chipid, mt_uint32 len, mt_uint8 *pframe)
{
	mt_uint8 *pbuf;
	mt_uint32 count;
	mt_uint32 llen = len + 2;
	if(llen > MT_MAX_BUFFERSIZE) {
		PRINTL(DBG_LVL_CHIP,"frame_exchange: Frame too long - %i\n", llen);
	}
	if(llen < 3) {
		PRINTL(DBG_LVL_CHIP,"frame_exchange: Frame too short - %i\n", llen);
	}
	write_hpi_8(chipid, (llen >> 8 ) & 0xff);
	write_hpi_8(chipid, llen & 0xff);
	write_hpi_8(chipid, *pframe++);
	len--;
        
	// Wait for command taken
	count = 0x10000000;
	while(count--) {
		if(!check_response_rdy())
		break;
	}
	if(count == 0xffffffff) {
		PRINTL(DBG_LVL_CHIP,"frame_exchange: Frame tx stalled\n");
		return MT_CHIP_FAILED_TO_RESPOND;
	}
	while(len--)
		write_hpi_8(chipid, *pframe++);

	// Wait for response ready
/*	count = 0x10000000;
	while(count--) {
		if(check_response_rdy())
			break;
	}
	if(count == 0xffffffff) {
		if(debug_level & DBG_LVL_CHIP)
			PRINTL(DBG_LVL_CHIP,"frame_exchange: Response timeout\n");
		return MT_CHIP_FAILED_TO_RESPOND;
 	}
*/	
	pbuf = &chips[chipid].buf[0];
	pbuf[0] = read_hpi_8(chipid);
	pbuf[1] = read_hpi_8(chipid);
 	llen = (pbuf[0] << 8 ) | pbuf[1];
	pbuf += 2;
	while(llen-- > 2)
		*pbuf++ = read_hpi_8(chipid);
	
	return handle_rx_frame_chip(chipid);
}


mt_uint32 send_frame_chip(mt_uint32 chipid, mt_uint32 len, mt_uint8 *pframe)
{
	PRINTL(DBG_LVL_CHIP,"in send_frame_chip:\n");	
	return new_frame_exchange(chipid, len, pframe);
}


void rx_frame_chip(mt_uint32 chipid)
{
	mt_uint8 *pbuf;
	mt_uint32 llen;

	pbuf = &chips[chipid].buf[0];
	pbuf[0] = read_hpi_8(chipid);
	pbuf[1] = read_hpi_8(chipid);
	llen = (pbuf[0] << 8 ) | pbuf[1];
	pbuf += 2;
	while(llen-- > 2)
		*pbuf++ = read_hpi_8(chipid);
                                                                                
	handle_rx_frame_chip(chipid);
}


mt_uint32 tx_frame_chip(mt_uint32 chipid, mt_uint32 len, mt_uint8 *pframe)
{
	mt_uint32 llen = len + 2;
	if(llen > MT_MAX_BUFFERSIZE) {
		PRINTL(DBG_LVL_CHIP,"tx_frame_chip: Frame too long - %i\n", llen);
	}

	write_hpi_8(chipid, (llen >> 8 ) & 0xff);
	write_hpi_8(chipid, llen & 0xff);

	while(len--)
		write_hpi_8(chipid, *pframe++);
                                                                                
	return MT_OK;
}
