/****************************************************************************
* File:         $Workfile$
* Revision:     $Revision: 1.15.4.8.2.1 $
* Date:         $Date: 2010-05-13 08:20:48 $
*
* Overview:     
* Functions:
*
* Author:       Ian Tsao
* 
* -- Metanoia Copyright Notice --
*
* (C) Copyright 2005 Metanoia, Inc. All rights reserved.
* 
* Metanoia reserves the right to change specifications without notice. 
* Copyright 2005 Metanoia, Inc. All rights reserved. Preliminary Data-Sheet 
* indicates this product is in design and the specification may change. 
* Electrical parametrics have not been analyzed and are not specified. Do not 
* use this data sheet as a design reference document. Please contact Metanoia 
* for a current data sheet that may be used as a design reference. All 
* contents of this document are protected by copyright law and may not be 
* reproduced without the express written consent of Metanoia, Inc. Metanoia, 
* the Metanoia logo, and combinations thereof are trademarks of Metanoia, Inc. 
* Other product names used in this publication are for identification purposes 
* only and may be trademarks or registered trademarks of their respective 
* companies. The contents of this document are provided in connection with 
* Metanoia, Inc. products. Metanoia, Inc. has made best efforts to ensure that 
* the information contained herein is accurate and reliable. However, Metanoia, 
* Inc. makes no warranties, express or implied, as to the accuracy or 
* completeness of the contents of this publication and is providing this 
* publication 'as is'. Metanoia, Inc. reserves the right to make changes to 
* specifications and product descriptions at any time without notice, and to 
* discontinue or make changes to its products at any time without notice. 
* Metanoia, Inc. does not assume any liability arising out of the application 
* or use of any product or circuit, and specifically disclaims any and all 
* liability, including without limitation special, consequential, or 
* incidental damages.
* 
****************************************************************************/
#include "dsl.h"

#ifdef __KERNEL__
#include <linux/interrupt.h>
#include <linux/list.h>
#include <asm/bitops.h>

#ifdef KERNEL_2_6
#include <linux/timer.h>
#endif

#define SYNC_THRESHOLD 30
//#define CHECK_15M_BOUNDARY 1
#define CHECK_15M_BOUNDARY 15

struct timer_list timer1, timer2;
struct tm statistic_save_tm = {0, 0, 0, 0, 0, 0};
#endif


/**
 * \file 
 * \brief API set that metanoia provide
 */

/** This file is the main API that export to user.
 */

/**
 * Reset statistic buffer of each port
 * \param portid
 * 		port number
 */
void init_var_statistic(mt_uint32 portid)
{
	mt_uint8 i = portid;
	mt_uint8 j, k, l;
		
	port_info[i].status = MT_OPSTATE_DISABLE;
	// cur == -1 mean there is no data
	port_info[i].inv15_cur = -1;
	port_info[i].inv24_cur = -1;

	for(j=0;j<((MT_INV_MAX15-1)>>3)+1;j++) {
		port_info[i].inv15_exist[j]=0; 	
	}
	
	for(j=0;j<((MT_INV_MAX24-1)>>3)+1;j++) {
		port_info[i].inv24_exist[j]=0; 	
	}		
	
	// clear statistic infor
	for(j=0;j<MT_IDX_LINE;j++) {
		// init 15 mins history buffer
		for(k=0;k<MT_INV_MAX15;k++) {
			for(l=0;l<MT_LINE_SIZE15;l++) {
				//line15[i][j][k][l]=0xffffffff;
				line15[i][j][k][l]=0;
			}
		}
		// init 15 mins init history buffer
		for(k=0;k<MT_INV_MAX15;k++) {
			for(l=0;l<MT_LINEINIT_SIZE15;l++) {
				//lineinit15[i][j][k][l]=0xffffffff;
				lineinit15[i][j][k][l]=0;
			}
		}
		// init 24 hour history buffer
		for(k=0;k<MT_INV_MAX24;k++) {
			for(l=0;l<MT_LINE_SIZE24;l++) {
				//line24[i][j][k][l]=0xffffffff;
				line24[i][j][k][l]=0;
			}
		}			
		// init 24 hour init history buffer
		for(k=0;k<MT_INV_MAX24;k++) {
			for(l=0;l<MT_LINEINIT_SIZE24;l++) {
				//line24[i][j][k][l]=0xffffffff;
				lineinit24[i][j][k][l]=0;
			}
		}
	}
	
	for(j=0;j<MT_IDX_CHAN;j++) {
		// init 15 mins history buffer
		for(k=0;k<MT_INV_MAX15;k++) {
			for(l=0;l<MT_CHAN_SIZE15;l++) {
				//chan15[i][j][k][l]=0xffffffff;
				chan15[i][j][k][l]=0;
			}
		}
		// init 24 hour history buffer
		for(k=0;k<MT_INV_MAX24;k++) {
			for(l=0;l<MT_CHAN_SIZE24;l++) {
				//chan24[i][j][k][l]=0xffffffff;
				chan24[i][j][k][l]=0;
			}
		}			
	}

//yalee980519 init global value 
#ifdef TR069	
	for(i=0 ; i < MAX_PORTS ; i++ ) {
		g_ShowtimeLinkRetrain[i]=0;                    
		g_ShowtimeInitErrors[i]=0;                     
		g_ShowtimeInitTimeouts[i]=0;                   
		g_ShowtimeErroredSecs[i]=0;
		g_ShowtimeSeverelyErroredSecs[i]=0;            
		g_ShowtimeFECErrors[i]=0;
		g_ShowtimeATUCFECErrors[i]=0;
		g_ShowtimeCRCErrors[i]=0;
		g_ShowtimeATUCErroredSecs[i]= 0;
		g_ShowtimeATUCCRCErrors[i]=0;
		g_ShowtimeATUCSeverelyErroredSecs[i]=0;

		g_LastShowtimeLinkRetrain[i]=0;
		g_LastShowtimeInitErrors[i]=0;
		g_LastShowtimeInitTimeouts[i]=0;               
		g_LastShowtimeErroredSecs[i]=0;
		g_LastShowtimeSeverelyErroredSecs[i]=0;        
		g_LastShowtimeFECErrors[i]=0;
		g_LastShowtimeATUCFECErrors[i]=0;
		g_LastShowtimeCRCErrors[i]=0;
		g_LastShowtimeATUCErroredSecs[i] = 0;
		g_LastShowtimeATUCCRCErrors[i] = 0;
		g_LastShowtimeATUCSeverelyErroredSecs[i] = 0;

		g_init_ShowtimeLinkRetrain[i]=0;
		g_init_ShowtimeInitErrors[i]=0;
		g_init_ShowtimeInitTimeouts[i]=0;              
		g_init_ShowtimeErroredSecs[i]=0;
		g_init_ShowtimeSeverelyErroredSecs[i]=0;       
		g_init_ShowtimeFECErrors[i]=0;
		g_init_ShowtimeATUCFECErrors[i]=0;
		g_init_ShowtimeCRCErrors[i]=0;
		g_init_ShowtimeATUCErroredSecs[i]= 0;
		g_init_ShowtimeATUCCRCErrors[i]=0;
		g_init_ShowtimeATUCSeverelyErroredSecs[i]=0;
	}
#endif	//TR069
}

/**
 * Initial procedure, must be called before any vdsllib API
 */
mt_ret init_port_info(void)
{
	mt_uint32 i;

	for(i=0;i<MAX_PORTS;i++) {
		// init port info structure data
		init_var_statistic(i);
	}	
	
	return MT_OK;
}


// -------------------------- MIB  -----------------------
/**
 * GetPortMIB get any MIB value on this port
 * \param portid
 * 		port ID number
 * \param oid
 * 		MIB Object ID string
 * \param pvalue
 * 		memory location to receive the value of the MIB
 * 
 * \return
 * 	- 0 : success
 * 	- -DSL_ERR_NODATA : no this interval data
 *  - -DSL_ERR_PARAM : invalid parameter
 * \par Notice:
 * 
 */
mt_ret get_port_mib(mt_uint32 portid, mt_uint8 *loid, mt_uint8 *strvalue)
{
	mt_ret ret = MT_OK;
	mt_uint8 ibyte, ibit;
	mt_uint8 start;
	mt_uint32 value;
	mt_uint32 groupid, tableid, paramid, idx, interval;
	
	if(portid == MT_PORT_BroadCast) {
		PRINTL(DBG_LVL_LIB, "Invalid portid=%i\n", portid);
		return -DSL_ERR_NOCHIP;
	}

	dsl_er(check_param(MT_PAR_PORT, 0, portid));

	oid2gtip(loid, &groupid, &tableid, &paramid, &idx, &interval);
	
	/*dsl_er(check_param(MT_PAR_PORT, 0, portid));
	dsl_er(check_param(MT_PAR_GROUP, 0, groupid));
	dsl_er(check_param(MT_PAR_TABLE, groupid, tableid));
	dsl_er(check_param_idx(groupid, tableid, idx));*/
	
	if((interval >= 0) && (interval != 0xff)) {
		if(interval != 0) interval--;
		else interval = 0x7f;	//interval == 0
		PRINTL(DBG_LVL_LIB, "Interval=%i\n",interval);
		ibyte = interval/8;
		ibit = interval-ibyte*8;
		//get history data
		if(groupid == MT_MIB_XdslGroup) {
			//interval need to translate to interval_idx
			// interval_idx = inv15_cur-interval-1
			switch(tableid) {
				case MT_MIB_Xdsl2PMLineHist15MinTable:
					if((interval == 0x7f) && (paramid > 2) && (paramid <= 9)) {	//interval == 0, mapping to current table
						tableid = MT_MIB_Xdsl2PMLineCurrTable;
						if(paramid == 9) paramid = 2;
						else paramid = paramid + 1;
						sprintf(loid, "%d.%d.%d.%d", groupid, tableid, paramid, idx);
						goto gpm_read_from_hpi;
					}
					else {
						if(interval > MT_INV_MAX15) return -DSL_ERR_NOINTERVAL;
						if(idx > MT_IDX_LINE) return -DSL_ERR_NOPARAM;
						if((port_info[portid].inv15_exist[ibyte] & 1<<ibit) == 0) return -DSL_ERR_NOINTERVAL;
						else {
							if((port_info[portid].inv15_exist[(MT_INV_MAX15-1)/8+1] & 1<<7) == 0) {	//buffer not scrolled
								start = 0;											//queue front
								interval = port_info[portid].inv15_cur - interval;	//reverse the interval to meet RFC requirement, port_info[portid].inv15_cur is the queue rear, to count interval bucket in the queue, we use "queue front" plus "reversed user inputted interval(port_info[portid].inv15_cur - interval)"
							}
							else {	//buffer scrolled
								start = (port_info[portid].inv15_cur + 1) % MT_INV_MAX15;	//(port_info[portid].inv15_cur % MT_INV_MAX15) is the queue rear, i.e. ((port_info[portid].inv15_cur+1) % MT_INV_MAX15) is the queue front
								interval = MT_INV_MAX15 - (interval + 1);	//reverse the interval to meet RFC requirement, port_info[portid].inv15_cur is the queue rear, to count interval bucket in the queue, we use "queue front" plus "reversed user inputted interval(MT_INV_MAX15 - (interval + 1))"
							}
							value = line15[portid][idx-1][(interval+start)%MT_INV_MAX15][paramid-1];
						}
					}
					break;

				case MT_MIB_Xdsl2PMLineHist1DayTable:
					if((interval == 0x7f) && (paramid > 2) && (paramid <= 9)) {	//interval == 0, mapping to current table
						tableid = MT_MIB_Xdsl2PMLineCurrTable;
						if(paramid == 9) paramid = 10;
						else paramid = paramid + 9;
						sprintf(loid, "%d.%d.%d.%d", groupid, tableid, paramid, idx);
						goto gpm_read_from_hpi;
					}
					else {
						if(interval > MT_INV_MAX24) return -DSL_ERR_NOINTERVAL;
						if(idx > MT_IDX_LINE) return -DSL_ERR_NOPARAM;
						if((port_info[portid].inv24_exist[ibyte] & 1<<ibit) == 0) return -DSL_ERR_NOINTERVAL;
						else {
							if((port_info[portid].inv24_exist[(MT_INV_MAX24-1)/8+1] & 1<<5) == 0) {	//buffer not scrolled
								start = 0;
								interval = port_info[portid].inv24_cur - interval;	//reverse the interval to meet RFC requirement
							}
							else {	//buffer scrolled
								start = (port_info[portid].inv24_cur + 1) % MT_INV_MAX24;
								interval = MT_INV_MAX24 - (interval + 1);	//reverse the interval to meet RFC requirement
							}
							value = line24[portid][idx-1][(interval+start)%MT_INV_MAX24][paramid-1];
						}
					}
					break;

				case MT_MIB_Xdsl2PMLineInitHist15MinTable:
					if((interval == 0x7f) && (paramid > 2) && (paramid < 7)) {	//interval == 0, mapping to current table
						tableid = MT_MIB_Xdsl2PMLineInitCurrTable;
						paramid = paramid + 1;
						sprintf(loid, "%d.%d.%d.%d", groupid, tableid, paramid, idx);
						goto gpm_read_from_hpi;
					}
					else {
						if(interval > MT_INV_MAX15) return -DSL_ERR_NOINTERVAL;
						if(idx > MT_IDX_LINE) return -DSL_ERR_NOPARAM;
						if((port_info[portid].inv15_exist[ibyte] & 1<<ibit) == 0) return -DSL_ERR_NOINTERVAL;
						else {
							if((port_info[portid].inv15_exist[(MT_INV_MAX15-1)/8+1] & 1<<7) == 0) {	//buffer not scrolled
								start = 0;
								interval = port_info[portid].inv15_cur - interval;	//reverse the interval to meet RFC requirement, port_info[portid].inv15_cur is the most recent interval data, we use backward counting to count the RFC expected interval from port_info[portid].inv15_cur
							}
							else {	//buffer scrolled
								start = (port_info[portid].inv15_cur + 1) % MT_INV_MAX15;	//(port_info[portid].inv15_cur % MT_INV_MAX15) is the most recent interval data, i.e. ((port_info[portid].inv15_cur+1) % MT_INV_MAX15) is the last recent interval data and also is the starting of the interval queue
								interval = MT_INV_MAX15 - (interval + 1);	//reverse the interval to meet RFC requirement, because buffer scrolled, all interval buckets are filled with interval data, we use forward counting to count the most recent interval, that's why we use (MT_INV_MAX15 - (interval + 1)) instead of backward counting "interval"
							}
							value = lineinit15[portid][idx-1][(interval+start)%MT_INV_MAX15][paramid-1];
						}
					}
					break;

				case MT_MIB_Xdsl2PMLineInitHist1DayTable:				
					if((interval == 0x7f) && (paramid > 2) && (paramid < 7)) {	//interval == 0, mapping to current table
						tableid = MT_MIB_Xdsl2PMLineInitCurrTable;
						paramid = paramid + 8;
						sprintf(loid, "%d.%d.%d.%d", groupid, tableid, paramid, idx);
						goto gpm_read_from_hpi;
					}
					else {
						if(interval > MT_INV_MAX24) return -DSL_ERR_NOINTERVAL;
						if(idx > MT_IDX_LINE) return -DSL_ERR_NOPARAM;
						if((port_info[portid].inv24_exist[ibyte] & 1<<ibit) == 0) return -DSL_ERR_NOINTERVAL;
						else {
							if((port_info[portid].inv24_exist[(MT_INV_MAX24-1)/8+1] & 1<<5) == 0) {	//buffer not scrolled
								start = 0;
								interval = port_info[portid].inv24_cur - interval;	//reverse the interval to meet RFC requirement
							}
							else {	//buffer scrolled
								start = (port_info[portid].inv24_cur + 1) % MT_INV_MAX24;
								interval = MT_INV_MAX24 - (interval + 1);	//reverse the interval to meet RFC requirement
							}
							value = lineinit24[portid][idx-1][(interval+start)%MT_INV_MAX24][paramid-1];
						}
					}
					break;

				case MT_MIB_Xdsl2PMChHist15MinTable:
					if((interval == 0x7f) && (paramid > 3) && (paramid < 6)) {	//interval == 0, mapping to current table
						tableid = MT_MIB_Xdsl2PMChCurrTable;
						paramid = paramid + 1;
						sprintf(loid, "%d.%d.%d.%d", groupid, tableid, paramid, idx);
						goto gpm_read_from_hpi;
					}
					else {
						if(interval > MT_INV_MAX15) return -DSL_ERR_NOINTERVAL;
						if(idx > MT_IDX_CHAN) return -DSL_ERR_NOPARAM;
						if((port_info[portid].inv15_exist[ibyte] & 1<<ibit) == 0) return -DSL_ERR_NOINTERVAL;
						else {
							if((port_info[portid].inv15_exist[(MT_INV_MAX15-1)/8+1] & 1<<7) == 0) {	//buffer not scrolled
								start = 0;
								interval = port_info[portid].inv15_cur - interval;	//reverse the interval to meet RFC requirement
							}
							else {	//buffer scrolled
								start = (port_info[portid].inv15_cur + 1) % MT_INV_MAX15;
								interval = MT_INV_MAX15 - (interval + 1);	//reverse the interval to meet RFC requirement
							}
							value = chan15[portid][idx-1][(interval+start)%MT_INV_MAX15][paramid-1];
						}
					}
					break;

				case MT_MIB_Xdsl2PMChHist1DTable:
					if((interval == 0x7f) && (paramid > 3) && (paramid < 6)) {	//interval == 0, mapping to current table
						tableid = MT_MIB_Xdsl2PMChCurrTable;
						paramid = paramid + 6;
						sprintf(loid, "%d.%d.%d.%d", groupid, tableid, paramid, idx);
						goto gpm_read_from_hpi;
					}
					else {
						if(interval > MT_INV_MAX24) return -DSL_ERR_NOINTERVAL;
						if(idx > MT_IDX_CHAN) return -DSL_ERR_NOPARAM;
						if((port_info[portid].inv24_exist[ibyte] & 1<<ibit) == 0) return -DSL_ERR_NOINTERVAL;
						else {
							if((port_info[portid].inv24_exist[(MT_INV_MAX24-1)/8+1] & 1<<5) == 0) {	//buffer not scrolled
								start = 0;
								interval = port_info[portid].inv24_cur - interval;	//reverse the interval to meet RFC requirement
							}
							else {	//buffer scrolled
								start = (port_info[portid].inv24_cur + 1) % MT_INV_MAX24;
								interval = MT_INV_MAX24 - (interval + 1);	//reverse the interval to meet RFC requirement
							}
							value = chan24[portid][idx-1][(interval+start)%MT_INV_MAX24][paramid-1];
						}
					}
				 	break;

				default:
					return -DSL_ERR_NOTABLE;
			}
		}
		else {
			PRINTL(DBG_LVL_LIB, "Only XdslGroup has interval data, interval of other Group must = 0");
			return -DSL_ERR_NOPARAM;
		}
	
		strvalue[0] = 0x00;	// insert mibsize high
		strvalue[1] = 0x01;	// insert mibsize low
		sprintf(strvalue+2, "%06x", value);
		PRINTL(DBG_LVL_LIB, "get_port_mib return %d\n", value);
		ret = MT_OK;
	}
	else {	//interval == 0xff or user inputted loid has no interval
		//yalee
		if(groupid == MT_MIB_XdslMetanoiaGroup && tableid == 2) {	//Layer 3 Private MIBs
			switch(paramid) {
#ifdef TR069
				case 3:
					//xdsl2StatsShowtimeLinkRetrain xstr(2.2.3)
					value = g_ShowtimeLinkRetrain[portid];
					break;
				case 4:
					//xdsl2StatsShowtimeInitErrors xstr(2.2.4)	
					value = g_ShowtimeInitErrors[portid];
					break;
				case 5:
					//xdsl2StatsShowtimeInitTimeouts xstr(2.2.5)
					value = g_ShowtimeInitTimeouts[portid];

					break;
				case 7:
					//xdsl2StatsShowtimeErroredSecs xstr(2.2.7)
					value = g_ShowtimeErroredSecs[portid];
					break;
				case 8:
					//xdsl2StatsShowtimeATUCErroredSecs xstr(2.2.8)
					value = g_ShowtimeATUCErroredSecs[portid];
					break;
				case 9:
					//xdsl2StatsShowtimeSeverelyErroredSecs xstr(2.2.9)
					value = g_ShowtimeSeverelyErroredSecs[portid];
					break;
				case 10:
					//xdsl2StatsShowtimeATUCSeverelyErroredSecs xstr(2.2.10)
					value = g_ShowtimeATUCSeverelyErroredSecs[portid];
					break;
				case 11:
					//xdsl2StatsShowtimeFECErrors xstr(2.2.11)
					value = g_ShowtimeFECErrors[portid];
					break;
				case 12:
					//xdsl2StatsShowtimeATUCFECErrors xstr(2.2.12)
					value = g_ShowtimeATUCFECErrors[portid];
					break;
				case 13:
					//xdsl2StatsShowtimeCRCErrors xstr(2.2.13)
					value = g_ShowtimeCRCErrors[portid];
					break;
				case 14:
					//xdsl2StatsShowtimeATUCCRCErrors xstr(2.2.14)
					value = g_ShowtimeATUCCRCErrors[portid];
					break;

				case 17:
					//xdsl2StatsLastShowtimeLinkRetrain xstr(2.2.17)
					value = g_LastShowtimeLinkRetrain[portid];
					break;
				case 18:
					//xdsl2StatsLastShowtimeInitErrors xstr(2.2.18)
					value = g_LastShowtimeInitErrors[portid];
					break;
				case 19:
					//xdsl2StatsLastShowtimeInitTimeouts xstr(2.2.19)
					value = g_LastShowtimeInitTimeouts[portid];
					break;
				case 21:
					//xdsl2StatsLastShowtimeErroredSecs xstr(2.2.21)
					value = g_LastShowtimeErroredSecs[portid];
					break;
				case 22:
					//xdsl2StatsLastShowtimeATUCErroredSecs xstr(2.2.22)
					value = g_LastShowtimeATUCErroredSecs[portid];
					break;
				case 23:
					//xdsl2StatsLastShowtimeSeverelyErroredSecs xstr(2.2.23)
					value = g_LastShowtimeSeverelyErroredSecs[portid];
					break;
				case 24:
					//xdsl2StatsLastShowtimeATUCSeverelyErroredSecs xstr(2.2.24)
					value = g_LastShowtimeATUCSeverelyErroredSecs[portid];
					break;
				case 25:
					//xdsl2StatsLastShowtimeFECErrors xstr(2.2.25)
					value = g_LastShowtimeFECErrors[portid];
					break;
				case 26:
					//xdsl2StatsLastShowtimeATUCFECErrors xstr(2.2.26)
					value = g_LastShowtimeATUCFECErrors[portid];
					break;
				case 27:
					//xdsl2StatsLastShowtimeCRCErrors xstr(2.2.27)
					value = g_LastShowtimeCRCErrors[portid];
					break;
				case 28:
					//xdsl2StatsLastShowtimeATUCCRCErrors xstr(2.2.28)
					value = g_LastShowtimeATUCCRCErrors[portid];
					break;
#endif	//TR069
				default:
					return -DSL_ERR_NOPARAM;
			}

			strvalue[0] = 0x00;	// insert mibsize high
			strvalue[1] = 0x01;	// insert mibsize low
			sprintf(strvalue+2, "%06x", value);
			ret = MT_OK;
		}
		else {
gpm_read_from_hpi:
			while(1) {	//make sure we can gain the HPI access authority
				if(test_and_set_bit(0, &atomic_ops) == FALSE) {
					ret = get_port_data_logic(portid, loid, strvalue);
					clear_bit(0, &atomic_ops);
					break;
				}
				else {
					MYDELAY(50);
				}
			}
		}
	}

	return ret;
}

/**
 * SetPortMIB sets a parameter in the MIB for the selected port.
 * \param portid
 * 		port ID number
 * \param oid
 * 		MIB Object ID string
 * \param pvalue
 * 		memory location with the value to be set into the MIB
 * 
 * \return
 * 	- 0 : success
 * 	- < 0 : error code 
 * \par Notice:
 * 		some MIB can't be set randomly, in order to protect system, SetPortMIB will filter it. and response with error\n
 * 		SetPortMIB don't support interval paramter
 *  - filter table list: LineConfProfile table, AlarmConfProfile Table
 */
mt_ret set_port_mib(mt_uint32 portid, mt_uint8 *loid, mt_uint8 *strvalue)
{
	mt_ret ret = MT_OK;

	dsl_er(check_param(MT_PAR_PORT, 0, portid));
		 
	while(1) {	//make sure we can gain the HPI access authority
		if(test_and_set_bit(0, &atomic_ops) == FALSE) {
			ret = set_port_data_logic(portid, loid, strvalue);
			clear_bit(0, &atomic_ops);
			break;
		}
		else {
			MYDELAY(50);
		}
	}

	return ret;
}

#ifndef CHIP_F4
/**
 * get MIBs array
 * \param portid
 * 		port ID number
 * \param tableid
* 		MIB table ID from vdslmib.h
 * \param idx
 * 		defined in MT_IDX_*
 * \param paramid
 * 		MIB parameter ID from vdslmib.h
 * \param interval
 * 		interval in perf, chan data which need interval\n
 *     set to 0 if this table don't need interval
 * \param count
 * 		paramid+count must not excess param count in the same table
 * \param value
 * 	    buffer to save values
 * \par Notice:
 * 		caller must make sure value array size must bigger than count*sizeof(mt_uint32)
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code
 */
mt_ret get_port_mib_array(mt_uint32 portid,mt_uint32 groupid,mt_uint32 tableid,mt_uint32 idx, 
							mt_uint32 paramid, mt_uint32 interval,mt_uint32 count,mt_uint32* value)
{
int i;

	dsl_er(check_param(MT_PAR_PORT,0,portid));
	dsl_er(check_param(MT_PAR_GROUP,0,groupid));
	dsl_er(check_param(MT_PAR_TABLE,0,tableid));	
	dsl_er(check_param_idx(groupid,tableid,idx));
		
	// need error checking
	// Need optimization
	
	
	for(i=0;i<count;i++)
	{
		dsl_er(get_port_mib(portid, groupid, tableid, idx, paramid+i, interval, value));
		value++;
	}
	return MT_OK;
}

/**
 * set MIBs array
 * \param portid
 * 		port ID number
 * \param idx
 * 		MT_IDX_* in vdsllib.h
 * \param mib_array
 * 		MIB array to be used
 * \param count_of_array
 * 		element number of specified MIB array
 * 
 * \return
 * 	- 0 : success
 * 	- < 0 : error code 
 */
mt_ret set_port_mib_array(mt_uint32 portid, mt_uint32 idx, mt_uint32 count_of_array, mib_value_set *mib_array)
{
	mt_uint32 i;
 	mt_ret ret;
	mt_uint32 groupid, tableid, paramid, value;

	dsl_er(check_param(MT_PAR_PORT,0,portid));
		 
 	for(i=0;i<count_of_array;i++)
 	{
		groupid=mib_array[i].group;
		tableid=mib_array[i].table;
		dsl_er(check_param_idx(groupid,tableid,idx));
		paramid=mib_array[i].param;
		value=mib_array[i].value;

		//PDINTD("group=%i, table=%i, param=%i, value=%i\n", groupid, tableid, paramid, value);

 		ret = set_port_data_logic(portid, groupid, tableid, idx, paramid, value);
 		if( ret != MT_OK ) return ret;
 	}
 	
 	return MT_OK;
}
#endif

/**
 * GetPhyStatus
 * \param phy_status
 * 	    defined in  MT_OPSTATE_*
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code
 */
mt_ret get_phy_status(mt_uint32 portid, mt_uint8 *phy_status, mt_uint8 *conn_status)
{
	mt_ret ret = MT_OK;
	mt_uint8 pdata[10];
	mt_uint32 value;

	dsl_er(check_param(MT_PAR_PORT,0,portid));
	
	if(port_info[portid].status == MT_OPSTATE_DISABLE) {
		if(phy_status != NULL) *phy_status = MT_OPSTATE_DISABLE;
		if(conn_status != NULL) *conn_status = connect_status;
		return MT_OK;
	}

	if(phy_status != NULL) {
		if(in_interrupt()) {	//request comes from interrupt handler
			*phy_status = port_info[portid].status;
		}
		else {
			while(1) {	//make sure we can gain the HPI access authority
				if(test_and_set_bit(0, &atomic_ops) == FALSE) {
					ret = get_port_data_logic(portid, xdslTerminal, pdata);
					if(ret == MT_OK) {
						value = simple_strtoul(pdata+2, NULL, 16);
						//Showtime: MT_MIB_noDefect=1, value= 1
						//Idle: MT_MIB_lossOfSignal=1, value= 110110
						//Training: MT_MIB_lossOfLink=1,MT_MIB_lossOfFraming=0,value= 110010
						PRINTL(DBG_LVL_LIB, "Phystatus = 0x%08x\n", value);
						value = (value >> MT_MIB_HandshakeIdle) & 0x00000003;
						switch(value) {
							case IDLE:
								*phy_status = MT_OPSTATE_IDLE;
								break;
							case HANDSHAKE:
								*phy_status = MT_OPSTATE_HANDSHAKE;
								break;
							case TRAINING:
								*phy_status = MT_OPSTATE_TRAINING;
								break;
							case SHOWTIME:
								*phy_status = MT_OPSTATE_SHOWTIME;
								break;
							default:
								PRINTL(DBG_LVL_LIB, "Impossible PhyStatus at Port%d!!\n", portid);
								break;
						}
					}
					else *phy_status = MT_OPSTATE_ERROR;

					clear_bit(0, &atomic_ops);
					break;
				}
				else {
					MYDELAY(50);
				}
			}
		}
	}

	if(conn_status != NULL) *conn_status = connect_status;

	return ret;
}

/**
 * get port's OpState
 * \param portid
 * 		port ID , start from 1
 * \param *state
 * 		return OpState defined in MT_OPSTATE_* to user
 * \return
 * 	- MT_OK: success
 * 	- < 0 : error code
 * 		
 */
mt_ret get_op_status(mt_uint32 portid, mt_uint8 *state)
{
	dsl_er(check_param(MT_PAR_PORT,0,portid));

	*state = port_info[portid].status;
	return MT_OK;
}


// ---------------------------- Conf Profile Management -------------------
#ifndef CHIP_F4
/**
 * Init profile variable to default value, run this function before modify profile variable
 * \param portid
 * 		port ID number
 * \param ps
 * 		profile content pointer
 * \par Detail
 * 		currently, profile id is the same with portid
 * 	
 */
mt_ret init_conf_profile(mt_uint32 portid,conf_profile_struct *ps)
{
	int i;

	dsl_er(check_param(MT_PAR_PORT,0,portid));

	/**
	 * Profile default configuration
	 * Detail value: 
	 */
	ps->conf.par.vdslLineConfProfileName=portid;

	for(i=0;i<45;i++)
		ps->conf.buf[i+1]=default_profile[i];

	return MT_OK;
}

/**
 * get profile setting by port
 * \param portid
 * 		port ID number
 * \param ps
 * 		profile content pointer
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code
 */
mt_ret get_conf_profile(mt_uint32 portid,conf_profile_struct *ps)
{
	int i;

	dsl_er(check_param(MT_PAR_PORT,0,portid));

	for(i=0;i<mib_table[MT_MIB_VdslLineConfProfileTable-1].entry_count;i++)
		get_port_data_logic(portid, MT_MIB_VdslGroup, MT_MIB_VdslLineConfProfileTable, MT_IDX_NEAR, i, &(ps->conf.buf[i]));

	return MT_OK;
}

/**
 * Setup profile by port
 * \param portid
 * 		port ID number
 * \param ps
 * 		profile content pointer
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code
 */
mt_ret set_conf_profile(mt_uint32 portid, conf_profile_struct *ps)
{
	int i;
	mt_uint32* buf;
	mt_ret ret;

	dsl_er(check_param(MT_PAR_PORT,0,portid));

	
	buf = ps->conf.buf;
	for(i=0;i<mib_table[MT_MIB_VdslLineConfProfileTable-1].entry_count;i++)
	{	
		PRINTL(DBG_LVL_LIB, "set_port_data_logic: portid=%i,param=%i,value=%i\n", portid, i+1, *(buf+i));
		ret = set_port_data_logic(portid, MT_MIB_VdslGroup, MT_MIB_VdslLineConfProfileTable, MT_IDX_NEAR, i+1, *(buf+i));		
		/*
		ret = get_port_data(portid, MT_MIB_VdslGroup, MT_MIB_VdslLineConfProfileTable, MT_IDX_NEAR, i+1, &value);				
		PRINTL(DBG_LVL_LIB, "Value=%i\n", value);
		*/
	}

	return MT_OK;	
}


// ---------------------------- Alarm Profile Management -------------------
/**
 * Setup Alarm Conf Profile default value
 * \param portid
 * 		port ID number
 * \param ps
 * 		profile content pointer
 */
mt_ret init_alarm_conf_profile(mt_uint32 portid,alarm_profile_struct *ps)
{
	/**
	 * Default Alarm Profile 
	 * Detail:
	 */

	int i;

	dsl_er(check_param(MT_PAR_PORT,0,portid));


	ps->conf.par.vdslLineAlarmConfProfileName=portid;
	 
	for(i=0;i<9;i++)
		ps->conf.buf[i+1]=default_alarm_profile[i];

	return MT_OK;	
    
}
	  	     
/**
 * Setup Alarm Profile
 * \param portid
 * 		port ID number
 * \param ps
 * 		profile content pointer 
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code
 */
mt_ret set_alarm_profile(mt_uint32 portid,alarm_profile_struct *ps)
{
	int i;
	mt_ret ret;
	mt_uint32* buf;
	buf = ps->conf.buf;

	dsl_er(check_param(MT_PAR_PORT,0,portid));

	
	for(i=0;i<mib_table[MT_MIB_VdslLineAlarmConfProfileTable-1].entry_count;i++)
		ret= set_port_data_logic(portid, MT_MIB_VdslGroup, MT_MIB_VdslLineAlarmConfProfileTable, MT_IDX_NEAR, i+1, *(buf+i));				

	return DSL_ERR_NODATA;
}

/**
 * Get Alarm Profile
 * \param portid
 * 		port ID number
 * \param ps
 * 		profile content pointer 
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code
 */
mt_ret get_alarm_profile(mt_uint32 portid,alarm_profile_struct *ps)
{
int i;
	dsl_er(check_param(MT_PAR_PORT,0,portid));

	
	for(i=0;i<mib_table[MT_MIB_VdslLineAlarmConfProfileTable-1].entry_count;i++)
		get_port_data_logic(portid, MT_MIB_VdslGroup, MT_MIB_VdslLineConfProfileTable, MT_IDX_NEAR, i+1, &(ps->conf.buf[i]));

	return DSL_ERR_NODATA;

}


// -------------------------- Notification -----------------------
/**
 * Setup Notification callback function or user space program process id
 * \param nparam
 * 		parameter which contained callback function or user space program process id
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code
 */
mt_ret setup_notification(notification_param nparam)
{
#ifdef __KERNEL__
	notification_link_t *tmp;

	if( nparam.nif == NULL && nparam.npid == 0 ) return DSL_ERR_NOPARAM;
	
	tmp = (notification_link_t *)kmalloc(sizeof(notification_link_t), GFP_KERNEL);
	if(!tmp){
		PRINTL(DBG_LVL_LIB, "setup_notification: could not allocate memory for notification interface\n");
		return DSL_ERR_NOTOK;
	}

	if( nparam.nif == NULL ){
		tmp->param.nif = NULL;
		tmp->param.npid = nparam.npid;
	}
	else{	// nparam.npid == 0
		tmp->param.nif = nparam.nif;
		tmp->param.npid = 0;
	}

	list_add(&tmp->link, &notification_linkq);
#else
	if( nparam.nif == NULL ) return DSL_ERR_NOPARAM;

	notification_intrfce = nparam.nif;
#endif	

	return MT_OK;
}

/**
 * Remove Notification callback function or user space program process id
 * \param nparam
 * 		parameter which contained callback function or user space program process id
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code
 */
mt_ret remove_notification(notification_param nparam)
{
#ifdef __KERNEL__
	struct	list_head *tmp;
	notification_link_t *qp = NULL;
	//int i = 0;

	if( nparam.nif == NULL && nparam.npid == 0 ) return DSL_ERR_NOPARAM;
	
	list_for_each(tmp, &notification_linkq){
		//PRINTD("%d \n", i++);

		qp=list_entry(tmp, notification_link_t, link);

		PRINTL(DBG_LVL_LIB, "qp->param.npid=%i\n", qp->param.npid);
		PRINTL(DBG_LVL_LIB, "nparam.npid=%i\n", nparam.npid);
		if(nparam.nif==NULL){
			if(nparam.npid==qp->param.npid)
				break;
		}
		else{	//nparam.nif!=NULL
			if(nparam.nif==qp->param.nif)
				break;
		}
	}

	if(qp){
		list_del(&qp->link);
		FREE(qp);
	}
#else
	if( nparam.nif == NULL ) return DSL_ERR_NOPARAM;

	if( notification_intrfce == nparam.nif )
		notification_intrfce = NULL;
	else
		PRINTL(DBG_LVL_LIB, "Invalid Notification Callback Function Pointer\n");
#endif

	return MT_OK;
}

/**
 * Give a mask to enable/disable some notification defined in VdslNotification
 * \param nid value defined in VdslNotifications
 * \notice example
 * 		mask_notification( )
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code 
 */
mt_ret mask_notification(mt_uint32 nid,mt_uint32 nstatus)
{
	masker_notification = nid;
	masker_status = nstatus;
	// Show new notification
	return MT_OK;	
}

/**
 * Process firmware notification status, and trap user, must be called every 500 ms
 * if there is any notification in each port, it will trap out once.
 * 
 */
mt_ret process_notification(void)
{
	mt_uint32 value, ret;
	mt_uint8 port, needed;
	notification_data ns;
#ifdef __KERNEL__
	struct task_struct *task;
	struct	list_head *tmp;
	notification_link_t	*qp;
#endif

	for(port=0;port<MAX_PORTS;port++)
	{
		needed = 0;

		PRINTL(DBG_LVL_LIB, "Process Notification for portid=%i before DISABLE\n",port);
		if((MT_AVAILABLE_PORT & (1<<port)) && (port_info[port].status == MT_OPSTATE_DISABLE)) continue;
		PRINTL(DBG_LVL_LIB, "Process Notification for portid=%i\n",port);
		
		ns.nstatus=0; ns.nid=0; ns.portid=port;
		// get firmware notification, MT_MIB_VdslMetanoiaGroup, MT_MIB_VdslNotifications
		ret = get_port_data_logic(port, MT_MIB_VdslMetanoiaGroup, MT_MIB_VdslMetanoiaTable, MT_IDX_NEAR, MT_MIB_VdslNotifications, &value);
		if(ret != MT_OK) return ret;
		if(masker_notification & value) // only trap user need
		{
			ns.nid = value;
			needed = 1;
		}
		
		// get firmware MT_MIB_VdslPhysCurrStatus
		ret = get_port_data_logic(port, MT_MIB_VdslGroup, MT_MIB_VdslPhysTable, MT_IDX_NEAR, MT_MIB_VdslPhysCurrStatus, &value);
		if(ret != MT_OK) return ret;		
		if( masker_status & value )
		{
			ns.nstatus = value;
			needed = 1;				
		}
		
		// trap only needed
		if( needed ){
#ifdef __KERNEL__
			list_for_each(tmp, &notification_linkq){
				qp=list_entry(tmp, notification_link_t, link);
				//PRINTD("notification_linkq=%i\n", notification_linkq);
				//PRINTD("qp->param.npid=%i\n", qp->param.npid);
				if(qp->param.nif==NULL){
					task=find_task_by_pid(qp->param.npid);
					send_sig(SIGUSR1, task, 1);
				}
				else	//qp->param.nif!=NULL
					(*(qp->param.nif))(&ns);
			}
#else
			(*notification_intrfce)(&ns);
#endif
		}
	}

	return MT_OK;
}
#endif


// -------------------------- statistic routine -----------------------
/**
 * This API must be called every 15 Min
 * \n get statistic data from firmware every 15 min
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code
 */
mt_ret mt_perf_15min(void)
{
	mt_ret ret = MT_OK;
	mt_uint8 ibyte,ibit;
	mt_uint8 loid_tmp[SMALL_BUF_SIZE];
	mt_uint32 i,j;
	mt_uint32* pvalue;
	mt_uint32 inv15_tmp;

	// save 15mins statistic data 
	for(i=0;i<MAX_PORTS;i++) {
		port_info[i].inv15_cur++;
		// count bucket indicator position
		inv15_tmp = port_info[i].inv15_cur % MT_INV_MAX15;
		
		ibyte = inv15_tmp / 8;
		ibit = inv15_tmp - ibyte*8;
		
		if(port_info[i].status == MT_OPSTATE_DISABLE) {
			port_info[i].inv15_exist[ibyte] &= ~(1 << ibit);	// indicate that this interval bucket does not exist 
		}
		else {
			port_info[i].inv15_exist[ibyte] |= 1 << ibit;	// indicate that this interval bucket exist
			// Line
			for(j=1;j<=MT_IDX_LINE;j++) {
				pvalue = line15[i][j-1][inv15_tmp];	// point to current interval bucket data buffer
				*pvalue++ = j;
				*pvalue++ = port_info[i].inv15_cur + 1;
				sprintf(loid_tmp, "%s.%d", xdsl2PMLHist15MMonitoredTime, j);
				ret = get_port_data_logic_value_array(i, loid_tmp, MT_LINE_SIZE15-2, pvalue);
				if( ret != MT_OK ) return ret;
			}
			// LineInit
			for(j=1;j<=MT_IDX_LINE;j++) {
				pvalue = lineinit15[i][j-1][inv15_tmp];	// point to current interval bucket data buffer
				*pvalue++ = port_info[i].inv15_cur + 1;
				sprintf(loid_tmp, "%s.%d", xdsl2PMLHistInit15MMonitoredTime, j);
				ret = get_port_data_logic_value_array(i, loid_tmp, MT_LINEINIT_SIZE15-1, pvalue);
				if( ret != MT_OK ) return ret;
			}
			// Chan
			for(j=1;j<=MT_IDX_CHAN;j++) {
				pvalue = chan15[i][j-1][inv15_tmp];
				*pvalue++ = j;
				*pvalue++ = port_info[i].inv15_cur + 1;
				sprintf(loid_tmp, "%s.%d", xdsl2PMChHist15MMonitoredTime, j);
				ret = get_port_data_logic_value_array(i, loid_tmp, MT_CHAN_SIZE15-2, pvalue);
				if( ret != MT_OK ) return ret;
			}
		}
	}

	return MT_OK;
}

/**
 * This API must be called every 24 Hour
 * \n get statistic data from firmware every 24 hour
 * \param portid
 * 		port ID number
 * \return 
 * 	- MT_OK: success
 * 	- < 0 : error code
 */
 
mt_ret mt_perf_24hour(void)
{
	mt_ret ret;
	mt_uint8 ibyte,ibit;
	mt_uint8 loid_tmp[SMALL_BUF_SIZE];
	mt_uint32 i,j;
	mt_uint32* pvalue;
	mt_uint32 inv24_tmp;

	// save 24hours statistic data 
	for(i=0;i<MAX_PORTS;i++) {
		port_info[i].inv24_cur++;
		inv24_tmp = port_info[i].inv24_cur % MT_INV_MAX24;
		
		ibyte = inv24_tmp / 8;
		ibit = inv24_tmp - ibyte*8;
		
		if(port_info[i].status == MT_OPSTATE_DISABLE) {
			port_info[i].inv24_exist[ibyte] &= ~(1 << ibit);
		}
		else {
			port_info[i].inv24_exist[ibyte] |= 1 << ibit;
			// Line
			for(j=1;j<=MT_IDX_LINE;j++) {
				pvalue = line24[i][j-1][inv24_tmp];
				*pvalue++ = j;
				*pvalue++ = port_info[i].inv24_cur + 1;
				sprintf(loid_tmp, "%s.%d", xdsl2PMLHist1DMonitoredTime, j);
				ret = get_port_data_logic_value_array(i, loid_tmp, MT_LINE_SIZE24-2, pvalue);
				if( ret != MT_OK ) return ret;
			}
			// LineInit
			for(j=1;j<=MT_IDX_LINE;j++) {
				pvalue = lineinit24[i][j-1][inv24_tmp];
				*pvalue++ = port_info[i].inv24_cur + 1;
				sprintf(loid_tmp, "%s.%d", xdsl2PMLHistInit1DMonitoredTime, j);
				ret = get_port_data_logic_value_array(i, loid_tmp, MT_LINEINIT_SIZE24-1, pvalue);
				if( ret != MT_OK ) return ret;
			}
			// Chan
			for(j=1;j<=MT_IDX_CHAN;j++) {
				pvalue = chan24[i][j-1][inv24_tmp];
				*pvalue++ = j;
				*pvalue++ = port_info[i].inv24_cur + 1;
				sprintf(loid_tmp, "%s.%d", xdsl2PMChHist1DMonitoredTime, j);
				ret = get_port_data_logic_value_array(i, loid_tmp, MT_CHAN_SIZE24-2, pvalue);
				if( ret != MT_OK ) return ret;
			}			
		}
	}

	return MT_OK;
}

//yalee980519
mt_ret mt_perf_showtime(void)
{
	mt_ret ret = MT_OK;
	mt_uint8 state;
	mt_uint32 value = 0;
	mt_uint32 portid, tmp_value;
	char strvalue[9];
#ifdef TR069
	static mt_uint8 vdsl_init_showtime_flag = 0;
#endif	//TR069
	//static mt_uint8 count = 0;
	
	/**yalee:one time of 5 sec**/
	/*if (count != 5) {
		count++;
		return MT_OK;
	}*/
	/**reset count**/
	//count = 0;

	for (portid = 0 ; portid < MAX_PORTS ; portid++)
	{
		get_op_status(portid, &state);
		if(state == MT_OPSTATE_DISABLE) {
			PRINTL(DBG_LVL_LIB, "Port%i disabled\n", portid);
			continue;
		}

		// vdsl2loid.h: vdslTerminal:2.1.5
		ret = gpdl_body(portid, xdslTerminal, strvalue);
		if(ret == MT_OK) {	//record port_status for interrupt handler
			value = simple_strtoul(strvalue+2, NULL, 16);
			//Showtime: MT_MIB_noDefect=1, value= 1
			//Idle: MT_MIB_lossOfSignal=1, value= 110110
			//Training: MT_MIB_lossOfLink=1,MT_MIB_lossOfFraming=0,value= 110010
			PRINTL(DBG_LVL_LIB, "Phystatus = 0x%08x\n", value);
			tmp_value = (value >> MT_MIB_HandshakeIdle) & 0x00000003;
			switch(tmp_value) {
				case IDLE:
					port_info[portid].status = MT_OPSTATE_IDLE;
					break;
				case HANDSHAKE:
					port_info[portid].status = MT_OPSTATE_HANDSHAKE;
					break;
				case TRAINING:
					port_info[portid].status = MT_OPSTATE_TRAINING;
					break;
				case SHOWTIME:
					port_info[portid].status = MT_OPSTATE_SHOWTIME;
					break;
				default:
					PRINTL(DBG_LVL_LIB, "Impossible PhyStatus at Port%d!!\n", portid);
					break;
			}
		}
		else {
			port_info[portid].status = MT_OPSTATE_ERROR;
			continue;
		}

#ifdef TR069
		if( value & 0x00000200 ) {	//showtime
		 	/*tr069:LOID
			ReceiveBlocks:soon
			TransmitBlocks:soon*/

			//**LinkRetrain:xdsl2PMLInitCurr1DayFullInits0**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLInitCurr1DayFullInits0, strvalue))
			{
					//**showtime_total = total - initi_showtime_total**//
				g_ShowtimeLinkRetrain[portid] = simple_strtoul(strvalue+2 , NULL , 16) - 
										g_init_ShowtimeLinkRetrain[portid];
			}

			//**InitErrors:xdsl2PMLInitCurr1DayFailedFullInits0**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLInitCurr1DayFailedFullInits0, strvalue))
			{
				g_ShowtimeInitErrors[portid] = simple_strtoul(strvalue+2 , NULL , 16) - 
										g_init_ShowtimeInitErrors[portid];
			}
			//**InitTimeouts:xdsl2PMLInitCurr1DayFailedShortInits0**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLInitCurr1DayFailedShortInits0, strvalue))
			{
				g_ShowtimeInitTimeouts[portid] = simple_strtoul(strvalue+2 , NULL , 16) - 
										g_init_ShowtimeInitTimeouts[portid];
			}
			//**ErroredSecs:xdsl2PMLCurr1DayEs1**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLCurr1DayEs1, strvalue))
			{
				g_ShowtimeErroredSecs[portid] = simple_strtoul(strvalue+2 , NULL , 16) -
										g_init_ShowtimeErroredSecs[portid];
			}
			//**ATUCErroredSecs:xdsl2PMLCurr1DayEs **//
			if (MT_OK == gpdl_body(portid, xdsl2PMLCurr1DayEs, strvalue))
			{
				g_ShowtimeATUCErroredSecs[portid] = simple_strtoul(strvalue+2 , NULL , 16) -
										g_init_ShowtimeATUCErroredSecs[portid];
			}
			//**SeverelyErroredSecs:xdsl2PMLCurr1DaySes1**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLCurr1DaySes1, strvalue))
			{
				g_ShowtimeSeverelyErroredSecs[portid] = simple_strtoul(strvalue+2 , NULL , 16) -
										g_init_ShowtimeSeverelyErroredSecs[portid];
			}
			//**ATUCSeverelyErroredSecs:xdsl2PMLCurr1DaySes //tr98**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLCurr1DaySes, strvalue))
			{
				g_ShowtimeATUCSeverelyErroredSecs[portid] = simple_strtoul(strvalue+2 , NULL , 16) -
										g_init_ShowtimeATUCSeverelyErroredSecs[portid];
			}
			//**FECErrors:xdsl2PMChCurr1DayCorrectedBlocks1**//
			if (MT_OK == gpdl_body(portid, xdsl2PMChCurr1DayCorrectedBlocks1, strvalue))
			{
				g_ShowtimeFECErrors[portid] = simple_strtoul(strvalue+2 , NULL , 16) - 
										g_init_ShowtimeFECErrors[portid];

			}
			//**ATUCFECErrors:xdsl2PMChCurr1DayCorrectedBlocks**//
			if (MT_OK == gpdl_body(portid, xdsl2PMChCurr1DayCorrectedBlocks, strvalue))
			{
				g_ShowtimeATUCFECErrors[portid] = simple_strtoul(strvalue+2 , NULL , 16) -
										g_init_ShowtimeATUCFECErrors[portid];
			}
			//**CRCrrors:xdsl2PMChCurr1DayCodingViolations1**//
			if (MT_OK == gpdl_body(portid, xdsl2PMChCurr1DayCodingViolations1, strvalue))
			{
				g_ShowtimeCRCErrors[portid] = simple_strtoul(strvalue+2 , NULL , 16) - 
										g_init_ShowtimeCRCErrors[portid];
			}
			//**ATUCCRCErrors:xdsl2PMChCurr1DayCodingViolations**//
			if (MT_OK == gpdl_body(portid, xdsl2PMChCurr1DayCodingViolations, strvalue))
			{
				g_ShowtimeATUCCRCErrors[portid] = simple_strtoul(strvalue+2 , NULL , 16) - 
										g_init_ShowtimeATUCCRCErrors[portid];
			}

			vdsl_init_showtime_flag = 1;

		}else{
			//Training log
			//**LinkRetrain:xdsl2PMLInitCurr1DayFullInits0**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLInitCurr1DayFullInits0, strvalue))
			{
					//**Traning**//
					g_init_ShowtimeLinkRetrain[portid] = simple_strtoul(strvalue+2 , NULL , 16);
			}
			//**InitErrors:xdsl2PMLInitCurr1DayFailedFullInits0**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLInitCurr1DayFailedFullInits0, strvalue))
			{
					g_init_ShowtimeInitErrors[portid] = simple_strtoul(strvalue+2 , NULL , 16);
			}
			//**InitTimeouts:xdsl2PMLInitCurr1DayFailedShortInits0**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLInitCurr1DayFailedShortInits0, strvalue))
			{
					g_init_ShowtimeInitTimeouts[portid] = simple_strtoul(strvalue+2 , NULL , 16);
			}
			//**ErroredSecs:xdsl2PMLCurr1DayEs1**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLCurr1DayEs1, strvalue))
			{
					g_init_ShowtimeErroredSecs[portid] = simple_strtoul(strvalue+2 , NULL , 16);
			}
			//**ATUCErroredSecs:xdsl2PMLCurr1DayEs **//
			if (MT_OK == gpdl_body(portid, xdsl2PMLCurr1DayEs, strvalue))
			{
					g_init_ShowtimeATUCErroredSecs[portid] = simple_strtoul(strvalue+2 , NULL , 16);
			}
			//**SeverelyErroredSecs:xdsl2PMLCurr1DaySes1**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLCurr1DaySes1, strvalue))
			{
					g_init_ShowtimeSeverelyErroredSecs[portid] = simple_strtoul(strvalue+2 , NULL , 16);
			}
			//**ATUCSeverelyErroredSecs:xdsl2PMLCurr1DaySes //tr98**//
			if (MT_OK == gpdl_body(portid, xdsl2PMLCurr1DaySes, strvalue))
			{
					g_init_ShowtimeSeverelyErroredSecs[portid] = simple_strtoul(strvalue+2 , NULL , 16);
			}
			//**FECErrors:xdsl2PMChCurr1DayCorrectedBlocks1**//
			if (MT_OK == gpdl_body(portid, xdsl2PMChCurr1DayCorrectedBlocks1, strvalue))
			{
					g_init_ShowtimeFECErrors[portid] = simple_strtoul(strvalue+2 , NULL , 16);

			}
			//**ATUCFECErrors:xdsl2PMChCurr1DayCorrectedBlocks**//
			if (MT_OK == gpdl_body(portid, xdsl2PMChCurr1DayCorrectedBlocks, strvalue))
			{
					g_init_ShowtimeATUCFECErrors[portid] = simple_strtoul(strvalue+2 , NULL , 16);
			}
			//**CRCrrors:xdsl2PMChCurr1DayCodingViolations1**//
			if (MT_OK == gpdl_body(portid, xdsl2PMChCurr1DayCodingViolations1, strvalue))
			{
					g_init_ShowtimeCRCErrors[portid] = simple_strtoul(strvalue+2 , NULL , 16);
			}
			//**ATUCCRCErrors:xdsl2PMChCurr1DayCodingViolations**//
			if (MT_OK == gpdl_body(portid, xdsl2PMChCurr1DayCodingViolations, strvalue))
			{
					g_init_ShowtimeCRCErrors[portid] = simple_strtoul(strvalue+2 , NULL , 16);
			}

			if(vdsl_init_showtime_flag == 1)
			{
				//ShowTime->Training
				g_LastShowtimeLinkRetrain[portid] 	= g_ShowtimeLinkRetrain[portid];
				g_LastShowtimeInitErrors[portid] 	= g_ShowtimeInitErrors[portid];
				g_LastShowtimeInitTimeouts[portid] 	= g_ShowtimeInitTimeouts[portid];
				g_LastShowtimeErroredSecs[portid] 	= g_ShowtimeErroredSecs[portid];
				g_LastShowtimeSeverelyErroredSecs[portid] = g_ShowtimeSeverelyErroredSecs[portid];
				g_LastShowtimeFECErrors[portid]		= g_ShowtimeFECErrors[portid];
				g_LastShowtimeATUCFECErrors[portid]	= g_ShowtimeATUCFECErrors[portid];
				g_LastShowtimeCRCErrors[portid]		= g_ShowtimeCRCErrors[portid];
				g_LastShowtimeATUCErroredSecs[portid] 	= g_ShowtimeATUCErroredSecs[portid] ;
				g_LastShowtimeATUCCRCErrors[portid] 	= g_ShowtimeATUCCRCErrors[portid];
				g_LastShowtimeATUCSeverelyErroredSecs[portid] = g_ShowtimeATUCSeverelyErroredSecs[portid];

			}

			vdsl_init_showtime_flag = 0;
		}
#endif	//TR069
	}

	return MT_OK;
}

/**
 * ClearPortStats clears port statistics
 * \param portid
 * 		port ID number 
 */
mt_ret clear_port_stats(mt_uint32 portid)
{
	// clear stats history
	init_var_statistic(portid);
	
	// need clear stats in firmware
	return MT_OK;
} 

/**
 * Statistics log trigger function
 */
mt_ret statistic_trigger(void)
{
	mt_ret ret;
	struct timeval tv;
	struct tm tm;
	int check_15min = 15;
	//int check_15min = 1;

	do_gettimeofday(&tv);
	ret = __offtime(&(tv.tv_sec), 0, &tm);
	if(ret == MT_OK) {
		if(tm.tm_min % check_15min == 1) {
			if(boottime.tm_min == tm.tm_min &&
				boottime.tm_hour == tm.tm_hour &&
				boottime.tm_mday == tm.tm_mday &&
				boottime.tm_mon == tm.tm_mon &&
				boottime.tm_year == tm.tm_year) {
				PRINTL(DBG_LVL_LIB, "statistic_timer_handler: boot skip\n");
				return MT_OK;
			}
			else {
				if(statistic_save_tm.tm_min == tm.tm_min &&
					statistic_save_tm.tm_hour == tm.tm_hour &&
					statistic_save_tm.tm_mday == tm.tm_mday &&
					statistic_save_tm.tm_mon == tm.tm_mon &&
					statistic_save_tm.tm_year == tm.tm_year) {//statistic has saved 30secs ago, so clear the statistic_save_flag this time
					PRINTL(DBG_LVL_LIB, "statistic_timer_handler: not save min:%d sec:%d\n", tm.tm_min, tm.tm_sec);
				}
				else {
					PRINTL(DBG_LVL_LIB, "statistic_timer_handler: save min:%d sec:%d\n", tm.tm_min, tm.tm_sec);
					statistic_save_tm.tm_min = tm.tm_min;
					statistic_save_tm.tm_hour = tm.tm_hour;
					statistic_save_tm.tm_mday = tm.tm_mday;
					statistic_save_tm.tm_mon = tm.tm_mon;
					statistic_save_tm.tm_year = tm.tm_year;

					mt_perf_15min();	//15mins save

					if(tm.tm_hour == 0 && tm.tm_min == 1)
						mt_perf_24hour();	//24hours save
				}
			}
		}
	}
	else
		PRINTD("statistic_timer_handler: time transform failed!\n");

	return ret;
}

#ifdef __KERNEL__
#ifndef CHIP_F4
void activate_timer1(void)	//EOC channel timer
{
	init_timer(&timer1);
	// "(64*HZ - 16*HZ)/1024 + jiffies"
	// equals to "0.046875 * HZ + jiffies"
	// approximates to "0.05 * HZ + jiffies"
	// HZ equals to 100 in ARM CPU
	timer1.expires = (((HZ<<6) - (HZ<<4))>>10) + jiffies;
	// set timer1 expires to 200ms
	//timer1.expires = 100 + jiffies;
	timer1.data = EOC_COMM_PORT;
	timer1.function = eoc_timer_handler;
	add_timer(&timer1);
	//schedule();
}

void deactivate_timer1(void)
{
	//del_timer(&timer1);
	del_timer_sync(&timer1);
}

void eoc_timer_handler(unsigned long chipid)
{
	struct task_struct *task;
	notification_link_t	*qp;

	activate_timer1();

	if(!list_empty(&notification_linkq)){
		//"dslman -t" will register as the first node of notification_linkq,
		//so use this method to obtain the corresponding pid
		qp=list_entry(notification_linkq.next, notification_link_t, link);
		task=find_task_by_pid(qp->param.npid);
		send_sig(SIGUSR2, task, 1);
	}
}

void activate_timer2(void)	//Statistic timer
{
	init_timer(&timer2);
	// HZ equals to 100 in ARM CPU
	// set timer2 expires to 1secs
	timer2.expires = (1*HZ) + jiffies;
	timer2.data = 0;
	timer2.function = statistic_timer_handler;
	add_timer(&timer2);
	//schedule();
}

void deactivate_timer2(void)
{
	//del_timer(&timer2);
	del_timer_sync(&timer2);
}
#endif	//CHIP_F4

int statistic_thread_handler(void *dummy)
{
	mt_ret ret;
	struct timeval tv;
	struct tm tm;
	unsigned long jiffies_tmp;

#ifdef KERNEL_2_6
	daemonize("vdsl_statistic");
	allow_signal(SIGKILL);
#else
	daemonize();
#endif

	PRINTL(DBG_LVL_LIB, "In statistic_thread_handler\n");

	while(!signal_pending(current)) {
		do_gettimeofday(&tv);
		ret = __offtime(&(tv.tv_sec), 0, &tm);
		if(ret == MT_OK) {
			if((tm.tm_min % CHECK_15M_BOUNDARY == 0) && (tm.tm_sec > SYNC_THRESHOLD)) {
				if(boottime.tm_min == tm.tm_min &&
					boottime.tm_hour == tm.tm_hour &&
					boottime.tm_mday == tm.tm_mday &&
					boottime.tm_mon == tm.tm_mon &&
					boottime.tm_year == tm.tm_year) {
					PRINTL(DBG_LVL_LIB, "statistic_timer_handler: boot skip\n");
					goto buttom_of_statistic;
				}
				else {
					if(statistic_save_tm.tm_min == tm.tm_min &&
						statistic_save_tm.tm_hour == tm.tm_hour &&
						statistic_save_tm.tm_mday == tm.tm_mday &&
						statistic_save_tm.tm_mon == tm.tm_mon &&
						statistic_save_tm.tm_year == tm.tm_year) {	//statistic has been saved 5 sec ago, so skip saving
						PRINTL(DBG_LVL_LIB, "statistic_timer_handler: not save min:%d sec:%d\n", tm.tm_min, tm.tm_sec);
					}
					else {
						if(test_and_set_bit(0, &atomic_ops) == FALSE) {
							PRINTL(DBG_LVL_LIB, "statistic_timer_handler: save min:%d sec:%d\n", tm.tm_min, tm.tm_sec);
							statistic_save_tm.tm_min = tm.tm_min;
							statistic_save_tm.tm_hour = tm.tm_hour;
							statistic_save_tm.tm_mday = tm.tm_mday;
							statistic_save_tm.tm_mon = tm.tm_mon;
							statistic_save_tm.tm_year = tm.tm_year;
		
							mt_perf_15min();	//15mins save
		
							if(tm.tm_hour == 0 && tm.tm_min == 0)
								mt_perf_24hour();	//24hours save
							
							clear_bit(0, &atomic_ops);
						}
						else
							PRINTL(DBG_LVL_LIB, "statistic_timer_handler: access violation\n");
					}
				}
			}
		}
		else
			PRINTD("statistic_timer_handler: time transform failed!\n");
	
		//PRINTD("y:%d m:%d d:%d h:%d m:%d s:%d\n", boottime.tm_year, boottime.tm_mon, boottime.tm_mday, boottime.tm_hour, boottime.tm_min, boottime.tm_sec);
		//PRINTD("y:%d m:%d d:%d h:%d m:%d s:%d\n", tm.tm_year, tm.tm_mon, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);

buttom_of_statistic:
//yalee980520
		if(test_and_set_bit(0, &atomic_ops) == FALSE) {
			mt_perf_showtime();
			//kernel_thread(mt_perf_showtime, NULL, NULL);
			clear_bit(0, &atomic_ops);
		}

		jiffies_tmp = jiffies;
		while(1) {
			MSLEEP(1000);
			if((jiffies - jiffies_tmp) >= (5 * HZ)) break;	//make sure the thread sleep enough for 5 seconds
		}
	}	
		
	complete_and_exit(&thread_exited, 1);
}

#ifndef KERNEL_2_6
/**
 *      msleep - sleep for a number of milliseconds
 *      @msecs: number of milliseconds to sleep
 *
 *      Issues schedule_timeout call for the specified number
 *      of milliseconds.
 *
 *      LOCKING:
 *      None.
 */
void mysleep(unsigned long msecs)
{
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(((HZ * msecs) / 1000) + 1);
}
#endif	//KERNEL_2_6
#endif	//__KERNEL__

EXPORT_SYMBOL(get_phy_status);
EXPORT_SYMBOL(get_port_mib);

// -------------------------- Debug -----------------------
// -------------------------- Misc -----------------------
