/*
This file is a part of the Leadtek Flash-OFDM driver, driver version: f.ff.f1.31 (january_wip3)

Copyright  2010 Leadtek Reseach Inc.

This program is free software; you may redistribute it and/or modify it under the terms of the GNU General
Public Licence as published by the Free Software Foundation; either version 2 [or version 3 or any other
relevant GPL licence type] of the Licence, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
Please see the GNU General Public Licence for more details.
You should have received a copy of the GNU General Public Licence along with this program; if not, please
write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.

Leadtek Research Inc.
18F, NO. 166, CHIEN-YI ROAD, CHUNG-HO CI
TAIPEE  HSIEN
Taiwan (23511), R. O. C
*/

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/usb.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include "ft1000_usb.h"

#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/netdevice.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/system.h>
#include <linux/kthread.h>

MODULE_DESCRIPTION("FT1000 EXPRESS CARD DRIVER");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE("Leadtek FT1000 Express Cards");


struct net_device *net_dev = NULL; //Chad
void *pFileStart;
ULONG FileLength;

#define VENDOR_ID 0x1291   /* Qualcomm vendor id */
#define PRODUCT_ID 0x11  /* fake product id */

/* table of devices that work with this driver */
static struct usb_device_id id_table[] = {
    {USB_DEVICE(VENDOR_ID, PRODUCT_ID) },
    { },
};

MODULE_DEVICE_TABLE (usb, id_table);

extern  struct ft1000_device *pdevobj[MAX_NUM_CARDS+2];

char *getfw (char *fn, int *pimgsz);

int ft1000_close(struct net_device *net);
void dsp_reload (struct ft1000_device *ft1000dev);
u16 init_ft1000_netdev(struct ft1000_device *ft1000dev, struct usb_interface *intf, int cardNum);
u16 reg_ft1000_netdev(struct ft1000_device *ft1000dev, struct usb_interface *intf);
int ft1000_poll(void* dev_id);
void ft1000_DestroyDevice(struct net_device *dev);
u16 ft1000_read_dpram16(struct ft1000_device *ft1000dev, USHORT indx, PUCHAR buffer, u8 highlow);
u16 ft1000_read_register(struct ft1000_device *ft1000dev, short* Data, u16 nRegIndx);
BOOLEAN gPollingfailed = FALSE;
int gCardState[8] = {0, 0, 0, 0, 0, 0, 0, 0};
int gCardIndex;
int gTotalCardNum;

int ft1000_poll_thread(void *arg);
//int ft1000_download_thread(void *arg);
int ft1000_download_thread(struct usb_interface *interface); //Chad

//For proc File
void ft1000InitProc(struct net_device *dev);
void ft1000CleanupProc(struct net_device *dev);; 

int ft1000_poll_thread(void *arg)
{
    int ret = STATUS_SUCCESS;
    struct ft1000_device *dev = (struct ft1000_device *)arg;
    // FT1000_INFO *info = dev->net->priv;
    FT1000_INFO *info = netdev_priv(dev->net); //ivan 20.08.09
    //int count = 0; //unused

    while(!kthread_should_stop() )
    {
        if ( info->disconnect )
            break;
        msleep_interruptible(50);
        //DEBUG("Polling...\n");
        if ( ! info->pollingFailed )
        {
            ret = ft1000_poll(arg); 
            if ( ret != STATUS_SUCCESS )
            {
                DEBUG("ft1000_poll_thread: polling failed\n");
                info->pollingFailed = TRUE;
            }
        }
        else
        {
            //may check card again to see if the card comes back, set pollingFailed to FALSE if card is good
            info->pollingFailed = FALSE;
            if ( info->disconnect || info->CardReady == 0 )
                break;
        }
    }

    info->pollingFailed = FALSE;
    DEBUG("returned from polling thread\n");
    return STATUS_SUCCESS;    
}

//int ft1000_download_thread(void *arg) //Chad
int ft1000_download_thread(struct usb_interface *interface) //Chad
{
    struct usb_host_interface *iface_desc;
    struct usb_endpoint_descriptor *endpoint;
    struct usb_device *dev;
    unsigned numaltsetting;    
    int i;
    struct ft1000_device *ft1000dev;
    FT1000_INFO *pft1000info;
    int cardNum;

    //struct usb_interface *interface = (struct usb_interface*)arg;	 //Chad

    if(!(ft1000dev = kmalloc(sizeof(struct ft1000_device), GFP_ATOMIC)))
    {
        printk("out of memory allocating device structure\n");
        return 0;
    }

    memset(ft1000dev, 0, sizeof(*ft1000dev));


    for (i=0; i<MAX_CARD_NUM ; i++)
    {
        if (gCardState[i] == 0)
        {
            gCardState[i] = 2;
            break;
        }
    }

    cardNum = i;

    //get usb device
    dev = interface_to_usbdev(interface);
    //DEBUG("ft1000_probe: usb device descriptor info:\n");
    //DEBUG("ft1000_probe: number of configuration is %d\n", dev->descriptor.bNumConfigurations);

    ft1000dev->dev = dev;
    ft1000dev->status = 0;
    ft1000dev->net = NULL;
    //ft1000dev->device_lock = SPIN_LOCK_UNLOCKED;
    spin_lock_init(&ft1000dev->device_lock);
    ft1000dev->tx_urb = usb_alloc_urb(0, GFP_ATOMIC);
    ft1000dev->rx_urb = usb_alloc_urb(0, GFP_ATOMIC);


    DEBUG("[Card %d]: ft1000_probe is called\n", cardNum);
    numaltsetting = interface->num_altsetting;
    //DEBUG("ft1000_probe: number of alt settings is :%d\n",numaltsetting);     
    iface_desc = interface->cur_altsetting;
    //DEBUG("ft1000_probe: number of endpoints is %d\n", iface_desc->desc.bNumEndpoints); 
    //DEBUG("ft1000_probe: descriptor type is %d\n", iface_desc->desc.bDescriptorType);
    //DEBUG("ft1000_probe: interface number is %d\n", iface_desc->desc.bInterfaceNumber);
    //DEBUG("ft1000_probe: alternatesetting is %d\n", iface_desc->desc.bAlternateSetting);
    //DEBUG("ft1000_probe: interface class is %d\n", iface_desc->desc.bInterfaceClass);
    //DEBUG("ft1000_probe: control endpoint info:\n");
    //DEBUG("ft1000_probe: descriptor0 type -- %d\n", iface_desc->endpoint[0].desc.bmAttributes);
    //DEBUG("ft1000_probe: descriptor1 type -- %d\n", iface_desc->endpoint[1].desc.bmAttributes);
    //DEBUG("ft1000_probe: descriptor2 type -- %d\n", iface_desc->endpoint[2].desc.bmAttributes);

    for (i=0; i< iface_desc->desc.bNumEndpoints;i++ )
    {
        endpoint = (struct usb_endpoint_descriptor *)&iface_desc->endpoint[i].desc;
        //DEBUG("endpoint %d\n", i);
        //DEBUG("bEndpointAddress=%x, bmAttributes=%x\n", endpoint->bEndpointAddress, endpoint->bmAttributes);
        if ( (endpoint->bEndpointAddress & USB_DIR_IN) && ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK))			 
        {
            ft1000dev->bulk_in_endpointAddr = endpoint->bEndpointAddress;
            //DEBUG("ft1000_probe: in: %d\n", endpoint->bEndpointAddress);
        }

        if (!(endpoint->bEndpointAddress & USB_DIR_IN) && ((endpoint->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK))
        {
            ft1000dev->bulk_out_endpointAddr = endpoint->bEndpointAddress;
            //DEBUG("ft1000_probe: out: %d\n", endpoint->bEndpointAddress);
        }
    }

    //DEBUG("bulk_in=%d, bulk_out=%d\n", ft1000dev->bulk_in_endpointAddr, ft1000dev->bulk_out_endpointAddr);


    //read DSP image
    pFileStart = (void*)getfw("/lib/modules/ft3000.img", &FileLength);

    if (pFileStart == NULL )
    {
        DEBUG ("[Card %d]: ft1000_probe: Read DSP image failed\n", cardNum);
        gCardState[cardNum] = 2;
        usb_set_intfdata(interface, NULL);
        //pft1000info->bDownloadThreadDone = TRUE;
        //pft1000info->bNetworkDeviceClosed = TRUE;
        return 0;
    }


    //download dsp image
    DEBUG("[Card %d]: ft1000_probe: initialized card...\n", cardNum);
    init_ft1000_netdev(ft1000dev, interface, cardNum);
    //pft1000info = ft1000dev->net->priv;
    pft1000info = netdev_priv(ft1000dev->net); //ivan 20.08.09
    //DEBUG("In probe: pft1000info=%x\n", pft1000info);
    //DEBUG("[Card %d]: ft1000_probe: start downloading dsp image...\n", cardNum);

    for (i=0; i<3 ; i++ )
    {
        if ( pft1000info-> disconnect != 1)
        {
            dsp_reload(ft1000dev);
            if (pft1000info->DSP_loading == 0)
            {
                break;
            }
            msleep(1000);
        }
        else
            break;
    }

    if ( pft1000info->DSP_loading )
    {
        DEBUG("[Card %d]: ERROR!!!! RETURN FROM ft1000_probe **********************\n", cardNum);
        gCardState[cardNum] = 2;
        pft1000info->bDownloadThreadDone = TRUE;
        pft1000info->bNetworkDeviceClosed = TRUE;
        return 0;
    }   

    pft1000info->pPollThread = kthread_run(ft1000_poll_thread, ft1000dev, "ft1000_poll");
    while (!pft1000info->CardReady && (pft1000info->disconnect!=1))
    {
        if ( pft1000info->pollingFailed )
        {
            //spin_lock(&ft1000dev->device_lock);
            if ( pft1000info->pPollThread )
            {
                kthread_stop(pft1000info->pPollThread );
            }
            //spin_unlock(&ft1000dev->device_lock);
            gCardState[cardNum] = 2;
            pft1000info->bDownloadThreadDone = TRUE;
            pft1000info->bNetworkDeviceClosed = TRUE;
            return 0;
        }
        msleep(100);
        DEBUG("[Card %d]: ft1000_probe::Waiting for Card Ready\n", cardNum);
    }

    if ( pft1000info->disconnect != 1)
    {
        //initialize network device
        DEBUG("[Card %d]: ft1000_probe::Card Ready!!!! Registering network device\n", cardNum);

        reg_ft1000_netdev(ft1000dev, interface);

        pft1000info->NetDevRegDone = 1;

        gCardState[cardNum] = 1;

        gTotalCardNum ++;
    }

    pft1000info->bDownloadThreadDone = TRUE;

    // Chad
    net_dev = ft1000dev->net;       
    ft1000InitProc(ft1000dev->net);                 
    if ( pft1000info->pPollThread ) {
        msleep(2500); //Chad: Must wait for connecting complete
        DEBUG("[Card %d]: Stopping Polling thread ...\n", pft1000info->CardNumber);
        kthread_stop(pft1000info->pPollThread );
    } 
    return 0;

}

//---------------------------------------------------------------------------
// Function:    ft1000_probe
//
// Parameters:  struct usb_interface *interface  - passed by USB core
//              struct usb_device_id *id         - passed by USB core
// Returns:     0 - success
//
// Description: This function is invoked when the express card is plugged in
//
// Notes:
//
//---------------------------------------------------------------------------
static int ft1000_probe(struct usb_interface *interface, const struct usb_device_id *id)
{

    DEBUG("Express card usb driver version : %x\n",  FT1000_DRV_VER);
    DEBUG("*** THIS IS THE JANUARY 2010 RELEASE ***\n");
    DEBUG("*** ALL BUGS ADDED BY IVAN KLIMEK ***\n");
    DEBUG("*** KLIMEK@CNL.SK ***\n");
    //kthread_run(ft1000_download_thread, interface, " ");  //Chad
    ft1000_download_thread(interface); //Chad
    return 0;
}

//---------------------------------------------------------------------------
// Function:    ft1000_disconnect
//
// Parameters:  struct usb_interface *interface  - passed by USB core
//              
// Returns:     0 - success
//
// Description: This function is invoked when the express card is plugged out
//
// Notes:
//
//---------------------------------------------------------------------------
static void ft1000_disconnect(struct usb_interface *interface)
{
    FT1000_INFO *pft1000info;



    pft1000info = (PFT1000_INFO)usb_get_intfdata(interface);

    //DEBUG("In disconnect pft1000info=%x\n", pft1000info);
    DEBUG("[Card %d]: ft1000_disconnect is called\n", pft1000info->CardNumber);


    if (pft1000info)
    {
    
                pft1000info->disconnect = 1;
                //sleep for some time, wait for all thread terminate
                //msleep(3000); //Chad
                //stop heartbeat thread
                if ( pft1000info->pHbThread )
                {
                    DEBUG("[Card %d]: Stopping heartbeat thread\n", pft1000info->CardNumber);
                    //kthread_stop(pft1000info->pHbThread);
                } 
		if (gCardState[pft1000info->CardNumber] == 1)
		{
		
			//spin_lock(&(pft1000info->pFt1000Dev->device_lock));
			if ( pft1000info->CardReady && pft1000info->pPollThread )
			{
				DEBUG ("[Card %d]: Stopping polling thread\n", pft1000info->CardNumber);
				//kthread_stop(pft1000info->pPollThread );
			}
			//spin_unlock(&(pft1000info->pFt1000Dev->device_lock));
			
			DEBUG("[Card %d]: ft1000_disconnect: threads are terminated\n", pft1000info->CardNumber);
		
			if (pft1000info->pFt1000Dev->net)
			{
                //Chad
                ft1000CleanupProc(pft1000info->pFt1000Dev->net);

				DEBUG("[Card %d]: ft1000_disconnect: destroy char driver\n", pft1000info->CardNumber);
				ft1000_DestroyDevice(pft1000info->pFt1000Dev->net);
				//DEBUG("ft1000_disconnect: calling ft1000_close\n");
				//ft1000_close(pft1000info->pFt1000Dev->net);
				//DEBUG("ft1000_disconnect: ft1000_close is called\n");
                                //DEBUG ("ft1000_disconnect: netdev->refcnt=%d\n", pft1000info->pFt1000Dev->net->refcnt);
				unregister_netdev(pft1000info->pFt1000Dev->net);
				DEBUG("[Card %d]: ft1000_disconnect: network device unregisterd\n", pft1000info->CardNumber);
				free_netdev(pft1000info->pFt1000Dev->net);

			}
			
			usb_free_urb(pft1000info->pFt1000Dev->rx_urb);
			usb_free_urb(pft1000info->pFt1000Dev->tx_urb);
		
			DEBUG("[Card %d]: ft1000_disconnect: urb freed\n", pft1000info->CardNumber);
			kfree (pft1000info->pFt1000Dev);
			gCardState[pft1000info->CardNumber] = 0;
		}
		else
		{
            //Chad
		    //while (!pft1000info->bDownloadThreadDone)
		    //    msleep(100);
		        
		    //case 1: char device is created
		    if (pft1000info->DeviceCreated)
		    {
		  DEBUG("[Card %d]: ft1000_disconnect: destroy char driver\n",pft1000info->CardNumber);
			ft1000_DestroyDevice(pft1000info->pFt1000Dev->net);
			unregister_netdev(pft1000info->pFt1000Dev->net);
			DEBUG("[Card %d]: ft1000_disconnect: network device unregisterd\n", pft1000info->CardNumber);
			free_netdev(pft1000info->pFt1000Dev->net);
			DEBUG("[Card %d]: ft1000disconnect: free FT1000DEV\n",pft1000info->CardNumber);
			kfree(pft1000info->pFt1000Dev);
			gCardState[pft1000info->CardNumber] = 0;
			goto exit;
		    }
		    
		    //case 2: char device is not created
		    if (pft1000info->pFt1000Dev)
		    {			
			if (pft1000info->pFt1000Dev->net)
			{
			   DEBUG("[Card %d]: ft1000disconnect: free net device\n",pft1000info->CardNumber);
                           free_netdev(pft1000info->pFt1000Dev->net);
			}
                        DEBUG("[Card %d]: ft1000disconnect: free FT1000DEV\n",pft1000info->CardNumber);
			kfree(pft1000info->pFt1000Dev);
			gCardState[pft1000info->CardNumber] = 0;
		    }
		}	
    }

    //terminate other kernel threads
    //in multiple instances case, first find the device 
    //in the link list
    /**if (pPollThread)
    {
        kthread_stop(pPollThread);
        DEBUG("Polling thread is killed \n");
    }**/

exit:	
   	gTotalCardNum --;
    return;
}

static struct usb_driver ft1000_usb_driver = {
    //.owner =    THIS_MODULE,
    .name  =    "ft1000usb",
    .probe =    ft1000_probe,
    .disconnect = ft1000_disconnect,
    .id_table = id_table,
};

//---------------------------------------------------------------------------
// Function:    usb_ft1000_init
//
// Parameters:  none
//              
// Returns:     0 - success
//
// Description: The entry point of the module, register the usb driver
//
// Notes:
//
//---------------------------------------------------------------------------
static int __init usb_ft1000_init(void)
{
    int ret = 0;
    
    DEBUG("Initialize and register the driver\n");

    ret = usb_register(&ft1000_usb_driver);
    if (ret)
        err("usb_register failed. Error number %d", ret);

    return ret;
}

//---------------------------------------------------------------------------
// Function:    usb_ft1000_exit
//
// Parameters:  
//              
// Returns:     
//
// Description: Moudle unload function, deregister usb driver
//
// Notes:
//
//---------------------------------------------------------------------------
static void __exit usb_ft1000_exit(void)
{        
    DEBUG("Deregister the driver\n");
    //Chad
    if (net_dev != NULL) {
        ft1000_close(net_dev);
        net_dev = NULL;
    }
    usb_deregister(&ft1000_usb_driver);
}

module_init (usb_ft1000_init);
module_exit (usb_ft1000_exit);


