#!/bin/sh

CNT_STATUS=`rdcsman 0x80355007 u32`
HOTPLUS_STATUS=`rdcsman 0x8035500f u32`
WIBRO_IF_IP=`rdcsman 0x80355002 u32`
WAN_STATUS=`rdcsman 0x80010007 u32`
BECEEM_WCM_SCRIPT_STARTING="/var/run/beceem_wcm_sh_is_starting"
STOP_BECEEM_WCM_SCRIPT="/var/run/stop_beceem_wcm_sh"
OPERATOR_ID="rdcsman 0x00355020 u32"
CONFIG_FILE_PATH="rdcsman 0x00355021 str"


. /usr/bin/scriptlib


csid()
{
	echo "hotplug_status: $(rdcsman 0x8035500f u32)"
	echo "wimax cnt_status: 0x$(printf "%x" $(rdcsman 0x80355007 u32))"
	echo "wan status: 0x$(printf "%x" $(rdcsman 0x80010007 u32))"
	echo "wimax if ip: $(rdcsman 0x80355002 ipv4)"
	echo "wimax or wibro: $(rdcsman 0x0035501c u32)"
	echo "name: $(rdcsman 0x00355011 str)"
	echo "identity: $(rdcsman 0x0035501b str)"
	echo "pass: $(rdcsman 0x00355012 str)"
	echo "BandwidthMHz: $(rdcsman 0x00355013 str)"
	echo "ScanMinFrequencyMHz: $(rdcsman 0x00355014 str)"
	echo "ScanMaxFrequencyMHz: $(rdcsman 0x00355015 str)"
	echo "ScanStepFrequencyMHz: $(rdcsman 0x00355016 str)"
	echo "Frequency 1: $(rdcsman 0x00355017 str)"
	echo "Frequency 2: $(rdcsman 0x00355018 str)"
	echo "Frequency 3: $(rdcsman 0x00355019 str)"
	echo "eap method: $(rdcsman 0x0035501a u32)"
	echo "authentication: $(rdcsman 0x0035501e u32)"
}

default()
{
	wrcsman    "0x00355011 \"0900605185@tatung.net.tw && \
				0x00355012 \"164047 && \
				0x0035501b \"0900605185@tatung.net.tw && \
				0x0035501d \"TATUNG && \
				0x00355013 \"10 && \
				0x00355014 \"0 && \
				0x00355015 \"0 && \
				0x00355016 \"0 && \
				0x00355017 \"2600 && \
				0x00355018 \"2610 && \
				0x00355019 \"2620 && \
				0x0035501a 0x00 && \
				0x0035501e 0x01 && \
				0x0035501c 0x01 \
			   "
}

beceem_start()
{
	if [ ! $HOTPLUS_STATUS -eq 2 ] ; then
		return 0
	fi

	if [ -e $STOP_BECEEM_WCM_SCRIPT ]; then
		logger -t beceem-wcm -s "Terminate beceem_start"
		return 1
	fi

	logger -t beceem-wcm -s "start"
	#uvm /usr/uo/wimaxd.conf.uyg.uo -o /etc/wimaxd.conf
	#wimaxd -c /etc/wimaxd.conf

	CONFIG_FILE_PATH=`rdcsman 0x00355021 str`
	if [ -e $CONFIG_FILE_PATH ]; then
		uvm $CONFIG_FILE_PATH -o /etc/wimaxd.conf
	else
		uvm /usr/uo/wimaxd.conf.uyg.uo -o /etc/wimaxd.conf
	fi

	#wimaxd -c /lib/wimaxd.conf
	wimaxd -c /etc/wimaxd.conf

	TIME=0
	TIMEOUT=120
	while [ true ] ; do
		if [ -e $STOP_BECEEM_WCM_SCRIPT ]; then
			logger -t beceem-wcm -s "Terminate beceem_start"
			return 1
		fi

		WAN_STATUS=`rdcsman 0x80010007 u32`	
		if [ $WAN_STATUS -eq 2 ] ; then
			# wan connected
			WIBRO_IF_IP=`rdcsman 0x80355002 u32`
			if [ $WIBRO_IF_IP -eq 0 ] ; then
				# get ip fail
				 wrcsman   "0x80355002 0x00 && \
							0x80355003 0x00 && \
							0x80355004 0x00 && \
							0x80355005 0x00 && \
							0x80355006 0x00 && \
							0x80355007 0x00 && \
							0x80355009 0x00 && \
							0x8035500a 0x00 && \
							0x801e0700 0x00 && \
							0x801e0e00 0x00 && \
							0x80010002 0x00 && \
							0x80010003 0x00 && \
							0x80010004 0x00 && \
							0x80010005 0x00 && \
							0x80010006 0x00"
				rmcsman     "0x801e1000"
				sleep 2
				logger -t beceem-wcm -s "get ip fail"
				return 1
			fi
			logger -t beceem-wcm -s "start ok"
			return 0
		else
			sleep 1
			TIME=$((TIME+1))
			if [ $TIME -gt $TIMEOUT ] ; then
				logger -t beceem-wcm -s "start error"
				return 1
			fi
		fi
	done
}

beceem_stop()
{
	logger -t beceem-wcm -s "stop"

	killall wimaxd
	killall udhcpc
	wrcsman 	"0x80355002 0x00 && \
				 0x80355003 0x00 && \
				 0x80355004 0x00 && \
				 0x80355005 0x00 && \
				 0x80355006 0x00 && \
				 0x80355007 0x00 && \
				 0x80355009 0x00 && \
				 0x8035500a 0x00 && \
				 0x801e0700 0x00 && \
				 0x801e0e00 0x00 && \
				 0x80010002 0x00 && \
				 0x80010003 0x00 && \
				 0x80010004 0x00 && \
				 0x80010005 0x00 && \
				 0x80010006 0x00"
	rmcsman     "0x801e1000"
	logger -t beceem-wcm -s "stop ok" 
	return 0
}


# main ##########################################################

case "$1" in

    start)
        beceem_start
        exit 0  
        ;;
    stop)
        beceem_stop
		exit 0
        ;;
	restart)
		stop
		start
		;;
	csid)
		csid
		exit 0
		;;
	default)
		default
		exit 0
		;;
    *)
        exit 1
        ;;
esac    

exit 0        

            
