#ifndef _SNOOPING_H
#define _SNOOPING_H

#include "igmprt.h"

#define MBLIST_BUF_SIZE 2048

/* igmp snooping table */
typedef struct _igmpsnp_source_t
{
	struct in_addr source;
	struct _igmpsnp_source_t *prev;
	struct _igmpsnp_source_t *next;
} igmpsnp_source_t;

typedef struct _igmpsnp_member_t
{
	struct in_addr member;
	int igmpsnp_mode;
	int igmpsnp_scount;
	igmpsnp_source_t  *igmpsnp_source;
	struct _igmpsnp_member_t *prev;
	struct _igmpsnp_member_t *next;
} igmpsnp_member_t;

typedef struct _igmpsnp_group_t
{
	struct in_addr group;
	int igmpsnp_mcount;
	igmpsnp_member_t *igmpsnp_member;
	struct _igmpsnp_group_t *prev;	
	struct _igmpsnp_group_t *next;	
} igmpsnp_group_t;

typedef struct _igmp_snoop_t 
{
	int igmpsnp_gcount;
	igmpsnp_group_t *igmpsnp_group;
} igmp_snoop_t;


/* snooping routines */
igmpsnp_group_t*
igmpsnp_group_lookup(igmpsnp_group_t *snp_gp, struct in_addr gp);

igmpsnp_member_t*
igmpsnp_member_lookup(igmpsnp_member_t *snp_mb, struct in_addr mb);

igmpsnp_source_t*
igmpsnp_source_lookup(igmpsnp_source_t *snp_src, struct in_addr src);

igmpsnp_group_t*
igmpsnp_group_create(struct in_addr gp);

igmpsnp_member_t*
igmpsnp_member_create(struct in_addr mb);

igmpsnp_source_t*
igmpsnp_source_create(struct in_addr src);

igmpsnp_group_t*
igmpsnp_group_add(igmp_snoop_t *snp, struct in_addr gp);

igmpsnp_member_t*
igmpsnp_member_add(igmpsnp_group_t *snp_gp, struct in_addr mb);

igmpsnp_source_t*
igmpsnp_source_add(igmpsnp_member_t *snp_mb, struct in_addr src);

void
igmpsnp_group_delete(igmp_snoop_t *snp, struct in_addr gp);

void
igmpsnp_member_delete(igmpsnp_group_t *snp_gp, struct in_addr mb);

void
igmpsnp_source_delete(igmpsnp_member_t *snp_mb, struct in_addr src);

void
igmpsnp_member_cleanup(igmpsnp_member_t *snp_mb);

void
igmpsnp_source_cleanup(igmpsnp_source_t *snp_src);

void
free_gp(igmp_snoop_t *snp, igmpsnp_group_t *snp_gp);

void
free_mb(igmpsnp_group_t *snp_gp, igmpsnp_member_t *snp_mb);

void
free_src(igmpsnp_member_t *snp_mb, igmpsnp_source_t *snp_src);

void 
maintain_snoop_table(void *pack);

void
igmpsnp_handle_report_v12(struct in_addr mb, igmpr_t *report);

void
igmpsnp_handle_report_v3(struct in_addr mb, void *pack);

void
igmpsnp_handle_isin(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src);

void
igmpsnp_handle_isex(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src);

void
igmpsnp_handle_toin(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src);
void
igmpsnp_handle_toex(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src);

void
igmpsnp_handle_allow(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src);

void
igmpsnp_handle_block(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src);

void 
igmpsnp_snoop_print(igmp_snoop_t *snp);

void 
igmpsnp_init(igmp_snoop_t *snp);

int
igmpsnp_member_list(struct in_addr gp, struct in_addr src, struct in_addr *mblist);

#endif // #ifndef _SNOOPING_H
