#include "snooping.h"

/* Table structure is shown below.
 
	snoop_table
	|
	-> igmpsnp_group[0] <-> igmpsnp_group[1] <-> ...<-> NULL
		 |
		 -> igmpsnp_member[0] <-> igmpsnp_member[1] <-> ... <-> NULL
		 		|
		 		-> igmpsnp_source[0] <-> igmpsnp_source[1] <-> ... <-> NULL
*/

igmp_snoop_t snoop_table;

igmpsnp_group_t*
igmpsnp_group_lookup(igmpsnp_group_t *snp_gp, struct in_addr gp)
{
	igmpsnp_group_t *gpix;
	
	for (gpix=snp_gp; gpix; gpix=gpix->next) {
		if (gpix->group.s_addr == gp.s_addr) {
			return gpix;
		}
	}
	
	return NULL;
}

igmpsnp_member_t*
igmpsnp_member_lookup(igmpsnp_member_t *snp_mb, struct in_addr mb)
{
	igmpsnp_member_t *mbix;
	
	for (mbix=snp_mb; mbix; mbix=mbix->next) {
		if (mbix->member.s_addr == mb.s_addr) {
			return mbix;
		}
	}
	
	return NULL;
}

igmpsnp_source_t*
igmpsnp_source_lookup(igmpsnp_source_t *snp_src, struct in_addr src)
{
	igmpsnp_source_t *srcix;
	
	for (srcix=snp_src; srcix; srcix=srcix->next) {
		if (srcix->source.s_addr == src.s_addr) {
			return srcix;
		}
	}
	
	return NULL;
}

igmpsnp_group_t*
igmpsnp_group_create(struct in_addr gp)
{
	igmpsnp_group_t *snp_gp;
	
	if ((snp_gp = (igmpsnp_group_t *)malloc(sizeof(igmpsnp_group_t)))) {
		snp_gp->group.s_addr   = gp.s_addr;
		snp_gp->igmpsnp_member = NULL;
		snp_gp->igmpsnp_mcount = 0;
		snp_gp->prev           = NULL;
		snp_gp->next           = NULL;
	}
	
	return snp_gp;
}

igmpsnp_member_t*
igmpsnp_member_create(struct in_addr mb)
{
	igmpsnp_member_t *snp_mb;
	
	if ((snp_mb = (igmpsnp_member_t *)malloc(sizeof(igmpsnp_member_t)))) {
		snp_mb->member.s_addr  = mb.s_addr;
		snp_mb->igmpsnp_source = NULL;
		snp_mb->igmpsnp_scount = 0;
		snp_mb->igmpsnp_mode   = IGMP_FMODE_INCLUDE;
		snp_mb->prev           = NULL;
		snp_mb->next           = NULL;
	}
	
	return snp_mb;
}

igmpsnp_source_t*
igmpsnp_source_create(struct in_addr src)
{
	igmpsnp_source_t *snp_src;
	
	if ((snp_src = (igmpsnp_source_t *)malloc(sizeof(igmpsnp_source_t)))) {
		snp_src->source.s_addr = src.s_addr;
		snp_src->prev          = NULL;
		snp_src->next          = NULL;
	}
	
	return snp_src;
}

igmpsnp_group_t*
igmpsnp_group_add(igmp_snoop_t *snp, struct in_addr gp)
{
	igmpsnp_group_t *snp_gp;
	
	assert(snp != NULL);
	if ((snp_gp = igmpsnp_group_lookup(snp->igmpsnp_group, gp)))
		return snp_gp;
	
	if ((snp_gp = igmpsnp_group_create(gp))) {
		snp_gp->next = snp->igmpsnp_group;
		snp->igmpsnp_group = snp_gp;
		if (snp_gp->next) {
			snp_gp->next->prev = snp_gp;
		}
		++snp->igmpsnp_gcount;
	}
	
	return snp_gp;
}

igmpsnp_member_t*
igmpsnp_member_add(igmpsnp_group_t *snp_gp, struct in_addr mb)
{
	igmpsnp_member_t *snp_mb;
	
	assert(snp_gp != NULL);
	if ((snp_mb = igmpsnp_member_lookup(snp_gp->igmpsnp_member, mb)))
		return snp_mb;
	
	if ((snp_mb = igmpsnp_member_create(mb))) {
		snp_mb->next = snp_gp->igmpsnp_member;
		snp_gp->igmpsnp_member = snp_mb;
		if (snp_mb->next) {
			snp_mb->next->prev = snp_mb;
		}
		++snp_gp->igmpsnp_mcount;
	}
	
	return snp_mb;
}

igmpsnp_source_t*
igmpsnp_source_add(igmpsnp_member_t *snp_mb, struct in_addr src)
{
	igmpsnp_source_t *snp_src;
	
	assert(snp_mb != NULL);
	if ((snp_src = igmpsnp_source_lookup(snp_mb->igmpsnp_source, src)))
		return snp_src;
	
	if ((snp_src = igmpsnp_source_create(src))) {
		snp_src->next = snp_mb->igmpsnp_source;
		snp_mb->igmpsnp_source = snp_src;
		if (snp_src->next) {
			snp_src->next->prev = snp_src;
		}
		++snp_mb->igmpsnp_scount;
	}
	
	return snp_src;
}

void
igmpsnp_group_delete(igmp_snoop_t *snp, struct in_addr gp)
{
	igmpsnp_group_t *snp_gp;
	igmpsnp_member_t *snp_mb, *mbix;
	igmpsnp_source_t *snp_src;
	
	assert(snp != NULL);
	if ((snp_gp = igmpsnp_group_lookup(snp->igmpsnp_group, gp))) {
		if (snp_gp->igmpsnp_member) {
			for (mbix=snp_gp->igmpsnp_member; mbix; mbix=mbix->next) {
				if (mbix->igmpsnp_source) {
					igmpsnp_source_cleanup(mbix->igmpsnp_source);
				}
			}
			igmpsnp_member_cleanup(snp_gp->igmpsnp_member);
		}
		free_gp(snp, snp_gp);
		--snp->igmpsnp_gcount;
	}
}

void
igmpsnp_member_delete(igmpsnp_group_t *snp_gp, struct in_addr mb)
{
	igmpsnp_member_t *snp_mb;
	
	assert(snp_gp != NULL);
	if ((snp_mb = igmpsnp_member_lookup(snp_gp->igmpsnp_member, mb))) {
		if (snp_mb->igmpsnp_source) {
			igmpsnp_source_cleanup(snp_mb->igmpsnp_source);
		}
		free_mb(snp_gp, snp_mb);
		--snp_gp->igmpsnp_mcount;
	}
}

void
igmpsnp_source_delete(igmpsnp_member_t *snp_mb, struct in_addr src)
{
	igmpsnp_source_t *snp_src;
	
	assert(snp_mb != NULL);
	if ((snp_src = igmpsnp_source_lookup(snp_mb->igmpsnp_source, src))) {
		free_src(snp_mb, snp_src);
		--snp_mb->igmpsnp_scount;
	}
}

void
igmpsnp_member_cleanup(igmpsnp_member_t *snp_mb)
{
	igmpsnp_member_t *mbix, *mbtmp;
	
	mbix = snp_mb;
	snp_mb = NULL;
	while (mbix) {
		mbtmp = mbix->next;
		free(mbix);
		mbix = mbtmp;
	}
}

void
igmpsnp_source_cleanup(igmpsnp_source_t *snp_src)
{
	igmpsnp_source_t *srcix, *srctmp;
	
	srcix = snp_src;
	snp_src = NULL;
	while (srcix) {
		srctmp = srcix->next;
		free(srcix);
		srcix = srctmp;
	}
}

void
free_gp(igmp_snoop_t *snp, igmpsnp_group_t *snp_gp)
{
	if (snp_gp->prev == NULL) { // Head Node
		snp->igmpsnp_group = snp_gp->next;
		if (snp->igmpsnp_group) {
			snp->igmpsnp_group->prev = NULL;
		}
	}
	else {
		if (snp_gp->next == NULL) {
			snp_gp->prev->next = NULL;
		}
		else {
			snp_gp->prev->next = snp_gp->next;
			snp_gp->next->prev = snp_gp->prev;
		}
	}
	free(snp_gp);
}

void
free_mb(igmpsnp_group_t *snp_gp, igmpsnp_member_t *snp_mb)
{
	if (snp_mb->prev == NULL) { // Head Node
		snp_gp->igmpsnp_member = snp_mb->next;
		if (snp_gp->igmpsnp_member) {
			snp_gp->igmpsnp_member->prev = NULL;
		}
	}
	else {
		if (snp_mb->next == NULL) {
			snp_mb->prev->next = NULL;
		}
		else {
			snp_mb->prev->next = snp_mb->next;
			snp_mb->next->prev = snp_mb->prev;
		}
	}
	free(snp_mb);
}

void
free_src(igmpsnp_member_t *snp_mb, igmpsnp_source_t *snp_src)
{
	if (snp_src->prev == NULL) { // Head Node
		snp_mb->igmpsnp_source = snp_src->next;
		if (snp_mb->igmpsnp_source) {
			snp_mb->igmpsnp_source->prev = NULL;
		}
	}
	else {
		if (snp_src->next == NULL) {
			snp_src->prev->next = NULL;
		}
		else {
			snp_src->prev->next = snp_src->next;
			snp_src->next->prev = snp_src->prev;
		}
	}
	free(snp_src);
}

void 
maintain_snoop_table(void *pack)
{
	struct ip *iphdr;
	unsigned char *igmprhdr;
	struct in_addr mb;
	int iphlen;
	extern unsigned long mask, subnet;
	
	igmpsnp_group_t *gpix, *gptmp;
	igmpsnp_member_t *mbix, *mbtmp;
	
	iphdr    = (struct ip *)pack;
	iphlen   = iphdr->ip_hl << 2;
	mb       = iphdr->ip_src;
	igmprhdr = pack + iphlen;
	
	// To judge report is coming from LAN side or not.
	if ((iphdr->ip_src.s_addr & mask) != subnet) {
		return;			
	}

	switch(*igmprhdr) {
		case IGMP_V1_MEMBERSHIP_REPORT:
		case IGMP_V2_MEMBERSHIP_REPORT:
			igmpsnp_handle_report_v12(mb, (igmpr_t*)igmprhdr);
			break;
			
		case IGMP_V3_MEMBERSHIP_REPORT:
		  igmpsnp_handle_report_v3(mb, (void*)igmprhdr);
      break;
      
		default:
			break;
	}
}

void
igmpsnp_handle_report_v12(struct in_addr mb, igmpr_t *report)
{
	struct in_addr gp, *src;
	
	gp  = report->igmpr_group;
	if (!IN_MULTICAST(ntohl(gp.s_addr)))
		return;

	igmpsnp_handle_isex(gp, mb, 0, NULL);
}

void
igmpsnp_handle_report_v3(struct in_addr mb, void *pack)
{
	igmp_report_t *report;
	struct in_addr gp, *src;
	u_short numsrc, numgrps;
	u_char rectype;
	int i;
	
	// parse igmpv3 report
	report = igmp_parse_pack(pack);
	
	numgrps = ntohs(report->igmpr_numgrps);
	for (i=0; i<numgrps; ++i) {
		gp = report->igmpr_group[i].igmpg_group;
		if (!IN_MULTICAST(ntohl(gp.s_addr)))
  	 	return;
	  
	  rectype = (u_char)report->igmpr_group[i].igmpg_type;
		numsrc  = ntohs(report->igmpr_group[i].igmpg_numsrc);
		src     = report->igmpr_group[i].igmpg_sources;
		switch(rectype) {
			case 1:
     		igmpsnp_handle_isin(gp, mb, numsrc, src);
     		break;
			case 2:
     		igmpsnp_handle_isex(gp, mb, numsrc, src);
     		break;
			case 3:
     		igmpsnp_handle_toin(gp, mb, numsrc, src);
     		break;
			case 4:
     		igmpsnp_handle_toex(gp, mb, numsrc, src);
     		break;
			case 5:
     		igmpsnp_handle_allow(gp, mb, numsrc, src);
     		break;
			case 6:
     		igmpsnp_handle_block(gp, mb, numsrc, src);
     		break;
			default:
     		break;
		}		
	}
}

void
igmpsnp_handle_isin(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src)
{
	igmpsnp_group_t *snp_gp;
	igmpsnp_member_t *snp_mb;
	igmpsnp_source_t *snp_src;
	int i;
	
	if ((snp_gp = igmpsnp_group_add(&snoop_table, gp)) == NULL)
		return;
	
	if ((snp_mb = igmpsnp_member_add(snp_gp, mb)) == NULL)
		return;
			
	/* include(A), isin(B) => include(A+B); 
		 exclude(Y), isin(A) => exclude(Y-A) */		
	if (snp_mb->igmpsnp_mode == IGMP_FMODE_INCLUDE) {
		for (i=0; i<numsrc; ++i) {
			igmpsnp_source_add(snp_mb, src[i]);
		}
	}
	else { // snp_mb->igmpsnp_mode == IGMP_FMODE_EXCLUDE
		for (i=0; i<numsrc; ++i) {
			igmpsnp_source_delete(snp_mb, src[i]);
		}
	}
}

void
igmpsnp_handle_isex(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src)
{
	igmpsnp_group_t *snp_gp;
	igmpsnp_member_t *snp_mb;
	igmpsnp_source_t *snp_src, *old_src;
	int i;
	
	if ((snp_gp = igmpsnp_group_add(&snoop_table, gp)) == NULL)
		return;
	
	if ((snp_mb = igmpsnp_member_add(snp_gp, mb)) == NULL)
		return;
	
	/* include(A), isex(B) => exclude(B-A); 
		 exclude(Y), isex(A) => exclude(Y*A) */
	old_src = snp_mb->igmpsnp_source;
	snp_mb->igmpsnp_source = NULL;
	snp_mb->igmpsnp_scount = 0;
	if (snp_mb->igmpsnp_mode == IGMP_FMODE_INCLUDE) {
		for (i=0; i<numsrc; ++i) {
			if ((snp_src = igmpsnp_source_lookup(old_src, src[i])) == NULL) {
				igmpsnp_source_add(snp_mb, src[i]);
			}
		}
		snp_mb->igmpsnp_mode = IGMP_FMODE_EXCLUDE;
	}
	else { // snp_mb->igmpsnp_mode == IGMP_FMODE_EXCLUDE
		for (i=0; i<numsrc; ++i) {
			if ((snp_src = igmpsnp_source_lookup(old_src, src[i])) != NULL) {
				igmpsnp_source_add(snp_mb, src[i]);
			}
		}
	}
	igmpsnp_source_cleanup(old_src);
}

void
igmpsnp_handle_toin(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src)
{
	/* include(A), isin(B) => include(A+B); 
		 exclude(Y), isin(A) => exclude(Y-A) */
	igmpsnp_handle_isin(gp, mb, numsrc, src);
}

void
igmpsnp_handle_toex(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src)
{
	/* include(A), isex(B) => exclude(B-A); 
		 exclude(Y), isex(A) => exclude(Y*A) */
	igmpsnp_handle_isex(gp, mb, numsrc, src);
}

void
igmpsnp_handle_allow(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src)
{
	/* include(A), allow(B) => include(A+B); 
		 exclude(Y), allow(A) => exclude(Y-A) */
	igmpsnp_handle_isin(gp, mb, numsrc, src);	
}

void
igmpsnp_handle_block(struct in_addr gp, struct in_addr mb, int numsrc, struct in_addr *src)
{
	igmpsnp_group_t *snp_gp;
	igmpsnp_member_t *snp_mb;
	igmpsnp_source_t *snp_src;
	int i;
	
	/* include(A), block(B) => include(A); 
		 exclude(Y), block(A) => exclude(Y) */
	//return;
		
	if ((snp_gp = igmpsnp_group_add(&snoop_table, gp)) == NULL)
		return;
	
	if ((snp_mb = igmpsnp_member_add(snp_gp, mb)) == NULL)
		return;
		
	if (snp_mb->igmpsnp_mode == IGMP_FMODE_INCLUDE) {
		for (i=0; i<numsrc; ++i) {
			if ((snp_src = igmpsnp_source_lookup(snp_mb->igmpsnp_source, src[i])) != NULL) {
				igmpsnp_source_delete(snp_mb, src[i]);
			}
		}
	}
	else { // snp_mb->igmpsnp_mode == IGMP_FMODE_EXCLUDE
		for (i=0; i<numsrc; ++i) {
			if ((snp_src = igmpsnp_source_lookup(snp_mb->igmpsnp_source, src[i])) == NULL) {
				igmpsnp_source_add(snp_mb, src[i]);
			}
		}
	}
}

void 
igmpsnp_table_print(igmp_snoop_t *snp)
{
	igmpsnp_group_t *gpix;
	igmpsnp_member_t *mbix;
	igmpsnp_source_t *srcix;


	printf("\nIGMPv3 Snooping Table\n");
	printf("\nGroup\t\t\tMember\t\t\tMode\tSource\n");
	printf("____________________________________________________________________________\n");
	for (gpix=snp->igmpsnp_group; gpix; gpix=gpix->next) {
		printf("%s(%d)\n", inet_ntoa(gpix->group), gpix->igmpsnp_mcount);
		for (mbix=gpix->igmpsnp_member; mbix; mbix=mbix->next) {
			printf("\t\t\t%5s(%d)\t%s\n", inet_ntoa(mbix->member), mbix->igmpsnp_scount, 
							mbix->igmpsnp_mode==IGMP_FMODE_INCLUDE?"INCL":"EXCL");
			for (srcix=mbix->igmpsnp_source; srcix; srcix=srcix->next) {
				printf("\t\t\t\t\t\t\t%15s\n", inet_ntoa(srcix->source));
			}
		}
	}
	printf("\n\n");
}

void 
igmpsnp_init(igmp_snoop_t *snp)
{
	snp->igmpsnp_gcount = 0;
	snp->igmpsnp_group  = NULL;
}

int
igmpsnp_member_list(struct in_addr gp, struct in_addr src, struct in_addr *mblist)
{
	igmpsnp_group_t *snp_gp;
	igmpsnp_source_t *snp_src;
	igmpsnp_member_t *mbix;
	int nmb = 0;
	
	if (mblist == NULL)
		return nmb;
	
	if ((snp_gp = igmpsnp_group_lookup(snoop_table.igmpsnp_group, gp)) == NULL)
		return nmb;
		
	for (mbix=snp_gp->igmpsnp_member; mbix; mbix=mbix->next) {
		if (mbix->igmpsnp_mode == IGMP_FMODE_INCLUDE) {
			if ((snp_src = igmpsnp_source_lookup(mbix->igmpsnp_source, src)) != NULL) {
				mblist[nmb++] = mbix->member;
			}
		}
		else { // snp_mb->igmpsnp_mode == IGMP_FMODE_EXCLUDE
			if ((snp_src = igmpsnp_source_lookup(mbix->igmpsnp_source, src)) == NULL) {
				mblist[nmb++] = mbix->member;
			}
		}
	}

	return nmb;
}
